import json
import math
import uuid
from datetime import datetime
from os import environ
from pathlib import Path
from typing import Any, Dict, List, Optional, Union, Callable, cast

# noinspection PyPep8Naming
import pyspark.sql.functions as F
from helix_fhir_client_sdk.filters.sort_field import SortField
from helix_fhir_client_sdk.function_types import RefreshTokenFunction
from helix_fhir_client_sdk.structures.get_access_token_result import (
    GetAccessTokenResult,
)
from pyspark import StorageLevel
from pyspark.ml.param import Param
from pyspark.rdd import RDD
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import col, from_json
from pyspark.sql.functions import explode
from pyspark.sql.types import (
    StringType,
    StructType,
    Row,
    DataType,
)

from spark_pipeline_framework.logger.log_level import LogLevel
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from spark_pipeline_framework.utilities.fhir_helpers.fhir_get_access_token import (
    fhir_get_access_token,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_receiver_exception import (
    FhirReceiverException,
)
from spark_pipeline_framework.transformers.fhir_receiver.v1.fhir_receiver_helpers import (
    FhirReceiverHelpers,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_get_response_schema import (
    FhirGetResponseSchema,
)
from spark_pipeline_framework.utilities.file_modes import FileWriteModes
from spark_pipeline_framework.utilities.pretty_print import get_pretty_data_frame
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
    sc,
)


class FhirReceiver(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        server_url: Optional[str] = None,
        resource: str,
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]],
        id_view: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        filter_by_resource: Optional[str] = None,
        filter_parameter: Optional[str] = None,
        additional_parameters: Optional[List[str]] = None,
        additional_parameters_view: Optional[str] = None,
        limit: Optional[int] = None,
        action: Optional[str] = None,
        action_payload: Optional[Dict[str, Any]] = None,
        include_only_properties: Optional[List[str]] = None,
        page_size: Optional[int] = None,
        batch_size: Optional[int] = None,
        last_updated_after: Optional[datetime] = None,
        last_updated_before: Optional[datetime] = None,
        sort_fields: Optional[List[SortField]] = None,
        auth_server_url: Optional[str] = None,
        auth_client_id: Optional[str] = None,
        auth_client_secret: Optional[str] = None,
        auth_login_token: Optional[str] = None,
        auth_scopes: Optional[List[str]] = None,
        separate_bundle_resources: bool = False,
        error_on_result_count: bool = True,
        expand_fhir_bundle: bool = True,
        accept_type: Optional[str] = None,
        content_type: Optional[str] = None,
        additional_request_headers: Optional[Dict[str, str]] = None,
        accept_encoding: Optional[str] = None,
        mode: str = FileWriteModes.MODE_OVERWRITE,
        ignore_status_codes: Optional[List[int]] = None,
        verify_counts_match: Optional[bool] = True,
        slug_column: Optional[str] = None,
        url_column: Optional[str] = None,
        error_view: Optional[str] = None,
        view: Optional[str] = None,
        retry_count: Optional[int] = None,
        exclude_status_codes_from_retry: Optional[List[int]] = None,
        num_partitions: Optional[int] = None,
        checkpoint_path: Optional[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = None,
        use_data_streaming: Optional[bool] = None,
        delta_lake_table: Optional[str] = None,
        schema: Optional[Union[StructType, DataType]] = None,
        cache_storage_level: Optional[StorageLevel] = None,
        graph_json: Optional[Dict[str, Any]] = None,
        run_synchronously: Optional[bool] = None,
        refresh_token_function: Optional[RefreshTokenFunction] = None,
    ) -> None:
        """
        Transformer to call and receive FHIR resources from a FHIR server


        :param server_url: server to call for FHIR
        :param resource: what FHIR resource to retrieve
        :param id_view: view that contains the ids to retrieve (if filter_by_resource is set then this is
                    used in the filter_by_resource query).  If not set, we load all the resources
        :param file_path: where to store the FHIR resources retrieved.  Can be either a path or a function
                    that returns path. Function is passed a loop number if inside a loop or else None
        :param name: name of transformer
        :param parameters: parameters
        :param progress_logger: progress logger
        :param auth_server_url: server url to call to get the authentication token
        :param filter_by_resource: filter the resource by this. e.g., /Condition?Patient=1
                (resource=Condition, filter_by_resource=Patient)
        :param filter_parameter: Instead of requesting ?patient=1,
                do ?subject:Patient=1 (if filter_parameter is subject)
        :param additional_parameters: Any additional parameters to send with request
        :param additional_parameters_view: (Optional) view name for additional parameters, e.g. search parameters
        :param auth_scopes: list of scopes to request permission for e.g., system/AllergyIntolerance.read
        :param limit: maximum number of resources to get
        :param action: (Optional) do an action e.g., $everything
        :param action_payload: (Optional) in case action needs a http request payload
        :param page_size: (Optional) use paging and get this many items in each page
        :param batch_size: (Optional) How many rows to have in each parallel partition to use for
                            retrieving from FHIR server
        :param last_updated_after: (Optional) Only get records newer than this
        :param last_updated_before: (Optional) Only get records older than this
        :param sort_fields: sort by fields in the resource
        :param separate_bundle_resources: True means we group the response by FHIR resources
                                            (check helix.fhir.client.sdk/tests/test_fhir_client_bundle.py)
        :param error_on_result_count: Throw exception when the response from FHIR count does not match the request count
        :param accept_type: (Optional) Accept header to use
        :param content_type: (Optional) Content-Type header to use
        :param additional_request_headers: (Optional) Additional request headers to use
                                            (Eg: {"Accept-Charset": "utf-8"})
        :param accept_encoding: (Optional) Accept-encoding header to use
        :param ignore_status_codes: (Optional) do not throw an exception for these HTTP status codes
        :param mode: if output files exist, should we overwrite or append
        :param slug_column: (Optional) use this column to set the security tags
        :param url_column: (Optional) column to read the url
        :param error_view: (Optional) log errors into this view (view only exists IF there are errors) and
                            don't throw exceptions.  schema: url, error_text, status_code
        :param use_data_streaming: (Optional) whether to use data streaming i.e., HTTP chunk transfer encoding
        :param delta_lake_table: whether to use delta lake for reading and writing files
        :param schema: the schema to apply after we receive the data
        :param cache_storage_level: (Optional) how to store the cache:
                                    https://sparkbyexamples.com/spark/spark-dataframe-cache-and-persist-explained/.
        :param graph_json: (Optional) a FHIR GraphDefinition resource to use for retrieving data
        :param run_synchronously: (Optional) Run on the Spark master to make debugging easier on dev machines
        :param refresh_token_function: (Optional) function to refresh the token
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert (
            isinstance(file_path, Path)
            or isinstance(file_path, str)
            or callable(file_path)
        ), type(file_path)

        assert not limit or isinstance(
            limit, int
        ), f"limit is not an int. it is a {type(limit)}"
        assert not batch_size or isinstance(
            batch_size, int
        ), f"batch_size is not an int. it is a {type(batch_size)}"
        assert not page_size or isinstance(
            page_size, int
        ), f"page_size is not an int. it is a {type(page_size)}"

        assert not last_updated_after or isinstance(
            last_updated_after, datetime
        ), f"last_updated_after is not a datetime. it is a {type(last_updated_after)}"

        assert not last_updated_before or isinstance(
            last_updated_before, datetime
        ), f"last_updated_before is not a datetime. it is a {type(last_updated_before)}"

        assert file_path

        self.logger = get_logger(__name__)

        self.server_url: Param[Optional[str]] = Param(self, "server_url", "")
        self._setDefault(server_url=None)

        self.resource: Param[str] = Param(self, "resource", "")
        self._setDefault(resource=None)

        self.filter_by_resource: Param[Optional[str]] = Param(
            self, "filter_by_resource", ""
        )
        self._setDefault(filter_by_resource=None)

        self.filter_parameter: Param[Optional[str]] = Param(
            self, "filter_parameter", ""
        )
        self._setDefault(filter_parameter=None)

        self.additional_parameters: Param[Optional[List[str]]] = Param(
            self, "additional_parameters", ""
        )
        self._setDefault(additional_parameters=None)

        self.additional_parameters_view: Param[Optional[str]] = Param(
            self, "additional_parameters_view", ""
        )
        self._setDefault(additional_parameters_view=None)

        self.id_view: Param[Optional[str]] = Param(self, "id_view", "")
        self._setDefault(id_view=None)

        self.file_path: Param[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.limit: Param[Optional[int]] = Param(self, "limit", "")
        self._setDefault(limit=None)

        self.action: Param[Optional[str]] = Param(self, "action", "")
        self._setDefault(action=None)

        self.action_payload: Param[Optional[Dict[str, Any]]] = Param(
            self, "action_payload", ""
        )
        self._setDefault(action_payload=None)

        self.include_only_properties: Param[Optional[List[str]]] = Param(
            self, "include_only_properties", ""
        )
        self._setDefault(include_only_properties=None)

        self.page_size: Param[Optional[int]] = Param(self, "page_size", "")
        self._setDefault(page_size=None)

        self.batch_size: Param[Optional[int]] = Param(self, "batch_size", "")
        self._setDefault(batch_size=None)

        self.last_updated_after: Param[Optional[datetime]] = Param(
            self, "last_updated_after", ""
        )
        self._setDefault(last_updated_after=None)

        self.last_updated_before: Param[Optional[datetime]] = Param(
            self, "last_updated_before", ""
        )
        self._setDefault(last_updated_before=None)

        self.sort_fields: Param[Optional[List[SortField]]] = Param(
            self, "sort_fields", ""
        )
        self._setDefault(sort_fields=None)

        self.auth_server_url: Param[Optional[str]] = Param(self, "auth_server_url", "")
        self._setDefault(auth_server_url=None)

        self.auth_client_id: Param[Optional[str]] = Param(self, "auth_client_id", "")
        self._setDefault(auth_client_id=None)

        self.auth_client_secret: Param[Optional[str]] = Param(
            self, "auth_client_secret", ""
        )
        self._setDefault(auth_client_secret=None)

        self.auth_login_token: Param[Optional[str]] = Param(
            self, "auth_login_token", ""
        )
        self._setDefault(auth_login_token=None)

        self.auth_scopes: Param[Optional[List[str]]] = Param(self, "auth_scopes", "")
        self._setDefault(auth_scopes=None)
        self.separate_bundle_resources: Param[bool] = Param(
            self, "separate_bundle_resources", ""
        )
        self._setDefault(separate_bundle_resources=False)

        self.expand_fhir_bundle: Param[bool] = Param(self, "expand_fhir_bundle", "")
        self._setDefault(expand_fhir_bundle=expand_fhir_bundle)

        self.error_on_result_count: Param[bool] = Param(
            self, "error_on_result_count", ""
        )
        self._setDefault(error_on_result_count=error_on_result_count)

        self.accept_type: Param[Optional[str]] = Param(self, "accept_type", "")
        self._setDefault(accept_type=accept_type)

        self.content_type: Param[Optional[str]] = Param(self, "content_type", "")
        self._setDefault(content_type=content_type)

        self.additional_request_headers: Param[Optional[Dict[str, str]]] = Param(
            self, "additional_request_headers", ""
        )
        self._setDefault(additional_request_headers=additional_request_headers)

        self.accept_encoding: Param[Optional[str]] = Param(self, "accept_encoding", "")
        self._setDefault(accept_encoding=accept_encoding)

        self.mode: Param[str] = Param(self, "mode", "")
        self._setDefault(mode=mode)

        self.ignore_status_codes: Param[Optional[List[int]]] = Param(
            self, "ignore_status_codes", ""
        )
        self._setDefault(ignore_status_codes=None)

        self.verify_counts_match: Param[Optional[bool]] = Param(
            self, "verify_counts_match", ""
        )
        self._setDefault(verify_counts_match=None)

        self.slug_column: Param[Optional[str]] = Param(self, "slug_column", "")
        self._setDefault(slug_column=None)

        self.url_column: Param[Optional[str]] = Param(self, "url_column", "")
        self._setDefault(url_column=None)

        self.error_view: Param[Optional[str]] = Param(self, "error_view", "")
        self._setDefault(error_view=None)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=None)

        self.retry_count: Param[Optional[int]] = Param(self, "retry_count", "")
        self._setDefault(retry_count=None)

        self.exclude_status_codes_from_retry: Param[Optional[List[int]]] = Param(
            self, "exclude_status_codes_from_retry", ""
        )
        self._setDefault(exclude_status_codes_from_retry=None)

        self.num_partitions: Param[Optional[int]] = Param(self, "num_partitions", "")
        self._setDefault(num_partitions=None)

        self.checkpoint_path: Param[
            Optional[Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]]
        ] = Param(self, "checkpoint_path", "")
        self._setDefault(checkpoint_path=None)

        self.use_data_streaming: Param[Optional[bool]] = Param(
            self, "use_data_streaming", ""
        )
        self._setDefault(use_data_streaming=None)

        self.delta_lake_table: Param[Optional[str]] = Param(
            self, "delta_lake_table", ""
        )
        self._setDefault(delta_lake_table=None)

        if delta_lake_table:
            assert schema is not None, f"schema must be supplied when using delta talk"

        self.schema: Param[Optional[Union[StructType, DataType]]] = Param(
            self, "schema", ""
        )
        self._setDefault(schema=None)

        self.cache_storage_level: Param[Optional[StorageLevel]] = Param(
            self, "cache_storage_level", ""
        )
        self._setDefault(cache_storage_level=None)

        self.graph_json: Param[Optional[Dict[str, Any]]] = Param(self, "graph_json", "")
        self._setDefault(graph_json=None)

        self.run_synchronously: Param[Optional[bool]] = Param(
            self, "run_synchronously", ""
        )
        self._setDefault(run_synchronously=run_synchronously)

        self.refresh_token_function: Param[Optional[RefreshTokenFunction]] = Param(
            self, "refresh_token_function", ""
        )
        self._setDefault(refresh_token_function=refresh_token_function)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        server_url: Optional[str] = self.getServerUrl()
        resource_name: str = self.getResource()
        parameters = self.getParameters()
        filter_by_resource: Optional[str] = self.getFilterByResource()
        filter_parameter: Optional[str] = self.getFilterParameter()
        additional_parameters: Optional[List[str]] = self.getAdditionalParameters()
        additional_parameters_view: Optional[str] = self.getOrDefault(
            self.additional_parameters_view
        )
        id_view: Optional[str] = self.getIdView()
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]] = (
            self.getFilePath()
        )
        if callable(file_path):
            file_path = file_path(self.loop_id)
        name: Optional[str] = self.getName()
        action: Optional[str] = self.getAction()
        action_payload: Optional[Dict[str, Any]] = self.getActionPayload()
        include_only_properties: Optional[List[str]] = self.getIncludeOnlyProperties()
        page_size: Optional[int] = self.getPageSize()
        batch_size: Optional[int] = self.getBatchSize()
        limit: Optional[int] = self.getLimit()
        last_updated_before: Optional[datetime] = self.getLastUpdatedBefore()
        last_updated_after: Optional[datetime] = self.getLastUpdateAfter()
        sort_fields: Optional[List[SortField]] = self.getSortFields()

        auth_server_url: Optional[str] = self.getAuthServerUrl()
        auth_client_id: Optional[str] = self.getAuthClientId()
        auth_client_secret: Optional[str] = self.getAuthClientSecret()
        auth_login_token: Optional[str] = self.getAuthLoginToken()
        auth_scopes: Optional[List[str]] = self.getAuthScopes()

        separate_bundle_resources: bool = self.getSeparateBundleResources()
        expand_fhir_bundle: bool = self.getExpandFhirBundle()

        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        auth_access_token: Optional[str] = None

        error_on_result_count: bool = self.getErrorOnResultCount()

        accept_type: Optional[str] = self.getAcceptType()
        content_type: Optional[str] = self.getContentType()
        accept_encoding: Optional[str] = self.getAcceptEncoding()
        additional_request_headers: Optional[Dict[str, str]] = (
            self.getAdditionalRequestHeaders()
        )

        ignore_status_codes: List[int] = self.getIgnoreStatusCodes() or []
        ignore_status_codes.append(200)

        verify_counts_match: Optional[bool] = self.getVerifyCountsMatch()
        mode: str = self.getMode()

        slug_column: Optional[str] = self.getOrDefault(self.slug_column)
        url_column: Optional[str] = self.getOrDefault(self.url_column)

        error_view: Optional[str] = self.getOrDefault(self.error_view)
        view: Optional[str] = self.getOrDefault(self.view)

        retry_count: Optional[int] = self.getOrDefault(self.retry_count)
        exclude_status_codes_from_retry: Optional[List[int]] = self.getOrDefault(
            self.exclude_status_codes_from_retry
        )

        num_partitions: Optional[int] = self.getOrDefault(self.num_partitions)

        checkpoint_path: Optional[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = self.getOrDefault(self.checkpoint_path)

        log_level: Optional[str] = environ.get("LOGLEVEL")

        use_data_streaming: Optional[bool] = self.getOrDefault(self.use_data_streaming)

        delta_lake_table: Optional[str] = self.getOrDefault(self.delta_lake_table)

        schema: Optional[Union[StructType, DataType]] = self.getOrDefault(self.schema)

        cache_storage_level: Optional[StorageLevel] = self.getOrDefault(
            self.cache_storage_level
        )

        graph_json: Optional[Dict[str, Any]] = self.getOrDefault(self.graph_json)

        run_synchronously: Optional[bool] = self.getOrDefault(self.run_synchronously)

        refresh_token_function: Optional[RefreshTokenFunction] = self.getOrDefault(
            self.refresh_token_function
        )

        if parameters and parameters.get("flow_name"):
            if additional_request_headers:
                additional_request_headers.update(
                    {
                        "Origin-Service": f"helix.pipelines:{parameters['flow_name']}",
                    }
                )
            else:
                additional_request_headers = {
                    "Origin-Service": f"helix.pipelines:{parameters['flow_name']}",
                }

        # get access token first so we can reuse it
        if auth_client_id and server_url:
            auth_access_token_result: GetAccessTokenResult = fhir_get_access_token(
                logger=self.logger,
                server_url=server_url,
                auth_server_url=auth_server_url,
                auth_client_id=auth_client_id,
                auth_client_secret=auth_client_secret,
                auth_login_token=auth_login_token,
                auth_scopes=auth_scopes,
                log_level=log_level,
            )
            auth_access_token = auth_access_token_result.access_token

        with ProgressLogMetric(
            name=f"{name}_fhir_receiver", progress_logger=progress_logger
        ):
            # if additional_parameters_view is given,
            # this is for generic use. e.g. to support blocking by adding search parameters in a query to FHIR
            if additional_parameters_view and additional_parameters:
                # created a df for additional_parameters_view
                # the data in this view MUST have 1 row per each parameter and only 1 column per each row.
                # e.g. "address-postalcode=10304,11040,11801" for a comma separate postal codes for Person resource
                df_additional_parameters_view: DataFrame = df.sparkSession.table(
                    additional_parameters_view
                )
                for row in df_additional_parameters_view.collect():
                    # get the data from each row for each parameter, e.g. "address-postalcode=10304,11040,11801"
                    parameter_values: str = str(row[0])
                    # add it to the additional_parameters
                    additional_parameters += [f"{parameter_values}"]

            # if we're calling for individual ids
            # noinspection GrazieInspection
            if id_view:
                id_df: DataFrame = df.sparkSession.table(id_view)
                if spark_is_data_frame_empty(df=id_df):
                    # nothing to do
                    return df

                assert "id" in id_df.columns

                has_token_col: bool = "token" in id_df.columns

                if limit and limit > 0:
                    id_df = id_df.limit(limit)

                row_count: int = id_df.count()

                self.logger.info(
                    f"----- Total {row_count} rows to request from {server_url or ''}/{resource_name}  -----"
                )
                desired_partitions: int = (
                    math.ceil(row_count / batch_size)
                    if batch_size and batch_size > 0
                    else 1
                )
                if not batch_size:  # if batching is disabled then call one at a time
                    desired_partitions = row_count
                self.logger.info(
                    f"----- Total Batches: {desired_partitions} for {server_url or ''}/{resource_name}  -----"
                )

                result_with_counts_and_responses: DataFrame
                if run_synchronously:
                    id_rows: List[Row] = id_df.collect()
                    result_rows: List[Row] = (
                        FhirReceiverHelpers.send_partition_request_to_server(
                            partition_index=0,
                            rows=id_rows,
                            batch_size=batch_size,
                            has_token_col=has_token_col,
                            server_url=server_url,
                            log_level=log_level,
                            action=action,
                            action_payload=action_payload,
                            additional_parameters=additional_parameters,
                            filter_by_resource=filter_by_resource,
                            filter_parameter=filter_parameter,
                            sort_fields=sort_fields,
                            auth_server_url=auth_server_url,
                            auth_client_id=auth_client_id,
                            auth_client_secret=auth_client_secret,
                            auth_login_token=auth_login_token,
                            auth_scopes=auth_scopes,
                            include_only_properties=include_only_properties,
                            separate_bundle_resources=separate_bundle_resources,
                            expand_fhir_bundle=expand_fhir_bundle,
                            accept_type=accept_type,
                            content_type=content_type,
                            additional_request_headers=additional_request_headers,
                            accept_encoding=accept_encoding,
                            slug_column=slug_column,
                            retry_count=retry_count,
                            exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                            limit=limit,
                            auth_access_token=auth_access_token,
                            resource_type=resource_name,
                            error_view=error_view,
                            url_column=url_column,
                            use_data_streaming=use_data_streaming,
                            graph_json=graph_json,
                        )
                    )
                    response_schema = FhirGetResponseSchema.get_schema()

                    result_with_counts_and_responses = df.sparkSession.createDataFrame(
                        result_rows, schema=response_schema
                    )
                else:
                    # run the above function on every partition
                    rdd: RDD[Row] = id_df.repartition(
                        desired_partitions
                    ).rdd.mapPartitionsWithIndex(
                        lambda partition_index, rows: FhirReceiverHelpers.send_partition_request_to_server(
                            partition_index=partition_index,
                            rows=rows,
                            batch_size=batch_size,
                            has_token_col=has_token_col,
                            server_url=server_url,
                            log_level=log_level,
                            action=action,
                            action_payload=action_payload,
                            additional_parameters=additional_parameters,
                            filter_by_resource=filter_by_resource,
                            filter_parameter=filter_parameter,
                            sort_fields=sort_fields,
                            auth_server_url=auth_server_url,
                            auth_client_id=auth_client_id,
                            auth_client_secret=auth_client_secret,
                            auth_login_token=auth_login_token,
                            auth_scopes=auth_scopes,
                            include_only_properties=include_only_properties,
                            separate_bundle_resources=separate_bundle_resources,
                            expand_fhir_bundle=expand_fhir_bundle,
                            accept_type=accept_type,
                            content_type=content_type,
                            additional_request_headers=additional_request_headers,
                            accept_encoding=accept_encoding,
                            slug_column=slug_column,
                            retry_count=retry_count,
                            exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                            limit=limit,
                            auth_access_token=auth_access_token,
                            resource_type=resource_name,
                            error_view=error_view,
                            url_column=url_column,
                            use_data_streaming=use_data_streaming,
                            graph_json=graph_json,
                        )
                    )

                    if has_token_col and not server_url:
                        assert slug_column
                        assert url_column
                        assert all(
                            [
                                c
                                for c in [url_column, slug_column, "resourceType"]
                                if [c in id_df.columns]
                            ]
                        )
                    response_schema = FhirGetResponseSchema.get_schema()

                    result_with_counts_and_responses = rdd.toDF(response_schema)

                # Now write to checkpoint if requested
                if checkpoint_path:
                    if callable(checkpoint_path):
                        checkpoint_path = checkpoint_path(self.loop_id)
                    checkpoint_file = (
                        f"{checkpoint_path}/{resource_name}/{uuid.uuid4()}"
                    )
                    if progress_logger:
                        progress_logger.write_to_log(
                            self.getName() or self.__class__.__name__,
                            f"Writing checkpoint to {checkpoint_file}",
                        )
                    checkpoint_file_format = "delta" if delta_lake_table else "parquet"
                    result_with_counts_and_responses.write.format(
                        checkpoint_file_format
                    ).save(checkpoint_file)
                    result_with_counts_and_responses = df.sparkSession.read.format(
                        checkpoint_file_format
                    ).load(checkpoint_file)
                else:
                    if cache_storage_level is None:
                        result_with_counts_and_responses = (
                            result_with_counts_and_responses.cache()
                        )
                    else:
                        result_with_counts_and_responses = (
                            result_with_counts_and_responses.persist(
                                storageLevel=cache_storage_level
                            )
                        )

                # don't need the large responses column to figure out if we had errors or not
                result_with_counts: DataFrame = result_with_counts_and_responses.select(
                    FhirGetResponseSchema.partition_index,
                    FhirGetResponseSchema.sent,
                    FhirGetResponseSchema.received,
                    FhirGetResponseSchema.first,
                    FhirGetResponseSchema.last,
                    FhirGetResponseSchema.error_text,
                    FhirGetResponseSchema.url,
                    FhirGetResponseSchema.status_code,
                    FhirGetResponseSchema.request_id,
                )
                result_with_counts = result_with_counts.cache()
                # find results that have a status code not in the ignore_status_codes
                results_with_counts_errored = result_with_counts.where(
                    (
                        (col(FhirGetResponseSchema.status_code).isNotNull())
                        & (
                            ~col(FhirGetResponseSchema.status_code).isin(
                                ignore_status_codes
                            )
                        )
                    )
                    | (col(FhirGetResponseSchema.error_text).isNotNull())
                )
                count_bad_requests: int = results_with_counts_errored.count()

                if count_bad_requests > 0:
                    result_with_counts.select(
                        FhirGetResponseSchema.partition_index,
                        FhirGetResponseSchema.sent,
                        FhirGetResponseSchema.received,
                        FhirGetResponseSchema.error_text,
                        FhirGetResponseSchema.status_code,
                        FhirGetResponseSchema.request_id,
                    ).show(truncate=False, n=1000000)
                    results_with_counts_errored.select(
                        FhirGetResponseSchema.partition_index,
                        FhirGetResponseSchema.sent,
                        FhirGetResponseSchema.received,
                        FhirGetResponseSchema.error_text,
                        FhirGetResponseSchema.url,
                        FhirGetResponseSchema.status_code,
                        FhirGetResponseSchema.request_id,
                    ).show(truncate=False, n=1000000)
                    first_error: str = (
                        results_with_counts_errored.select(
                            FhirGetResponseSchema.error_text
                        )
                        .limit(1)
                        .collect()[0][0]
                    )
                    first_url: str = (
                        results_with_counts_errored.select(FhirGetResponseSchema.url)
                        .limit(1)
                        .collect()[0][0]
                    )
                    first_error_status_code: Optional[int] = (
                        results_with_counts_errored.select(
                            FhirGetResponseSchema.status_code
                        )
                        .limit(1)
                        .collect()[0][0]
                    )
                    first_request_id: Optional[str] = (
                        results_with_counts_errored.select(
                            FhirGetResponseSchema.request_id
                        )
                        .limit(1)
                        .collect()[0][0]
                    )
                    if error_view:
                        errors_df = results_with_counts_errored.select(
                            FhirGetResponseSchema.url,
                            FhirGetResponseSchema.error_text,
                            FhirGetResponseSchema.status_code,
                            FhirGetResponseSchema.request_id,
                        )
                        errors_df.createOrReplaceTempView(error_view)
                        if progress_logger and not spark_is_data_frame_empty(errors_df):
                            progress_logger.log_event(
                                event_name="Errors receiving FHIR",
                                event_text=get_pretty_data_frame(
                                    df=errors_df,
                                    limit=100,
                                    name="Errors Receiving FHIR",
                                ),
                                log_level=LogLevel.INFO,
                            )
                        # filter out bad records
                        result_with_counts = result_with_counts.where(
                            ~(
                                (col(FhirGetResponseSchema.status_code).isNotNull())
                                & (
                                    ~col(FhirGetResponseSchema.status_code).isin(
                                        ignore_status_codes
                                    )
                                )
                            )
                            | (col(FhirGetResponseSchema.error_text).isNotNull())
                        )
                    else:
                        raise FhirReceiverException(
                            url=first_url,
                            response_text=first_error,
                            response_status_code=first_error_status_code,
                            message="Error receiving FHIR",
                            json_data=first_error,
                            request_id=first_request_id,
                        )

                if expand_fhir_bundle:
                    # noinspection PyUnresolvedReferences
                    count_sent_df: DataFrame = result_with_counts.agg(
                        F.sum(FhirGetResponseSchema.sent).alias(
                            FhirGetResponseSchema.sent
                        )
                    )
                    # count_sent_df.show()
                    # noinspection PyUnresolvedReferences
                    count_received_df: DataFrame = result_with_counts.agg(
                        F.sum(FhirGetResponseSchema.received).alias(
                            FhirGetResponseSchema.received
                        )
                    )
                    # count_received_df.show()
                    count_sent: Optional[int] = count_sent_df.collect()[0][0]
                    count_received: Optional[int] = count_received_df.collect()[0][0]
                    if (
                        (count_sent is not None and count_received is not None)
                        and (count_sent > count_received)
                        and error_on_result_count
                        and verify_counts_match
                    ):
                        result_with_counts.where(
                            col(FhirGetResponseSchema.sent)
                            != col(FhirGetResponseSchema.received)
                        ).select(
                            FhirGetResponseSchema.partition_index,
                            FhirGetResponseSchema.sent,
                            FhirGetResponseSchema.received,
                            FhirGetResponseSchema.error_text,
                            FhirGetResponseSchema.url,
                            FhirGetResponseSchema.request_id,
                        ).show(
                            truncate=False, n=1000000
                        )
                        first_url = (
                            result_with_counts.select(FhirGetResponseSchema.url)
                            .limit(1)
                            .collect()[0][0]
                        )
                        first_error_status_code = (
                            result_with_counts.select(FhirGetResponseSchema.status_code)
                            .limit(1)
                            .collect()[0][0]
                        )
                        first_request_id = (
                            result_with_counts.select(FhirGetResponseSchema.request_id)
                            .limit(1)
                            .collect()[0][0]
                        )
                        raise FhirReceiverException(
                            url=first_url,
                            response_text=None,
                            response_status_code=first_error_status_code,
                            message=f"Sent ({count_sent}) and Received ({count_received}) counts did not match",
                            json_data="",
                            request_id=first_request_id,
                        )
                    elif count_sent == count_received:
                        self.logger.info(
                            f"Sent ({count_sent}) and Received ({count_received}) counts matched "
                            + f"for {server_url or ''}/{resource_name}"
                        )
                    else:
                        self.logger.info(
                            f"Sent ({count_sent}) and Received ({count_received}) for "
                            + f"{server_url or ''}/{resource_name}"
                        )

                # turn list of list of string to list of strings
                # rdd1: RDD = rdd.flatMap(lambda a: a.responses)

                # noinspection PyUnresolvedReferences
                # result_df: DataFrame = rdd1.toDF(StringType())
                result_df: DataFrame = (
                    result_with_counts_and_responses.select(
                        explode(col(FhirGetResponseSchema.responses))
                    )
                    if expand_fhir_bundle
                    else result_with_counts_and_responses.select(
                        # col("responses")[0]["id"].alias("id"),
                        col(FhirGetResponseSchema.responses)[0].alias("bundle")
                    )
                )

                # result_df.printSchema()
                # TODO: remove any OperationOutcomes.  Some FHIR servers like to return these

                self.logger.info(
                    f"Executing requests and writing FHIR {resource_name} resources to {file_path}..."
                )
                if delta_lake_table:
                    if schema:
                        result_df = result_df.select(
                            from_json(
                                col("col"), schema=cast(StructType, schema)
                            ).alias("resource")
                        )
                        result_df = result_df.selectExpr("resource.*")
                    result_df.write.format("delta").mode(mode).save(str(file_path))
                    result_df = df.sparkSession.read.format("delta").load(
                        str(file_path)
                    )
                else:
                    result_df.write.format("text").mode(mode).save(str(file_path))
                    result_df = df.sparkSession.read.format("text").load(str(file_path))

                self.logger.info(
                    f"Received {result_df.count()} FHIR {resource_name} resources."
                )
                self.logger.info(
                    f"Reading from disk and counting rows for {resource_name}..."
                )
                file_row_count: int = result_df.count()
                self.logger.info(
                    f"Wrote {file_row_count} FHIR {resource_name} resources to {file_path}"
                )
                if progress_logger:
                    progress_logger.log_event(
                        event_name="Finished receiving FHIR",
                        event_text=json.dumps(
                            {
                                "message": f"Wrote {file_row_count} FHIR {resource_name} resources "
                                + f"to {file_path} (id view)",
                                "count": file_row_count,
                                "resourceType": resource_name,
                                "path": str(file_path),
                            },
                            default=str,
                        ),
                    )
                if view:
                    result_df.createOrReplaceTempView(view)
            else:  # get all resources
                result1 = FhirReceiverHelpers.get_batch_result(
                    page_size=page_size,
                    limit=limit,
                    server_url=server_url,
                    action=action,
                    action_payload=action_payload,
                    additional_parameters=additional_parameters,
                    filter_by_resource=filter_by_resource,
                    filter_parameter=filter_parameter,
                    resource_name=resource_name,
                    include_only_properties=include_only_properties,
                    last_updated_after=last_updated_after,
                    last_updated_before=last_updated_before,
                    sort_fields=sort_fields,
                    auth_server_url=auth_server_url,
                    auth_client_id=auth_client_id,
                    auth_client_secret=auth_client_secret,
                    auth_login_token=auth_login_token,
                    auth_access_token=auth_access_token,
                    auth_scopes=auth_scopes,
                    separate_bundle_resources=separate_bundle_resources,
                    expand_fhir_bundle=expand_fhir_bundle,
                    accept_type=accept_type,
                    content_type=content_type,
                    additional_request_headers=additional_request_headers,
                    accept_encoding=accept_encoding,
                    retry_count=retry_count,
                    exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                    log_level=log_level,
                    error_view=error_view,
                    ignore_status_codes=ignore_status_codes,
                    use_data_streaming=use_data_streaming,
                    graph_json=graph_json,
                    refresh_token_function=refresh_token_function,
                )
                resources = result1.resources
                errors = result1.errors
                rdd1: RDD[str] = (
                    sc(df).parallelize(resources, numSlices=num_partitions)
                    if num_partitions is not None
                    else sc(df).parallelize(resources)
                )

                list_df: DataFrame = rdd1.toDF(StringType())
                file_format = "delta" if delta_lake_table else "text"
                list_df.write.format(file_format).mode(mode).save(str(file_path))
                list_df = df.sparkSession.read.format(file_format).load(str(file_path))

                self.logger.info(f"Wrote FHIR data to {file_path}")

                if progress_logger:
                    progress_logger.log_event(
                        event_name="Finished receiving FHIR",
                        event_text=json.dumps(
                            {
                                "message": f"Wrote {list_df.count()} FHIR {resource_name} resources to "
                                + f"{file_path} (query)",
                                "count": list_df.count(),
                                "resourceType": resource_name,
                                "path": str(file_path),
                            },
                            default=str,
                        ),
                    )

                if view:
                    list_df.createOrReplaceTempView(view)
                if error_view:
                    errors_df = sc(df).parallelize(errors).toDF().cache()
                    errors_df.createOrReplaceTempView(error_view)
                    if progress_logger and not spark_is_data_frame_empty(errors_df):
                        progress_logger.log_event(
                            event_name="Errors receiving FHIR",
                            event_text=get_pretty_data_frame(
                                df=errors_df,
                                limit=100,
                                name="Errors Receiving FHIR",
                            ),
                            log_level=LogLevel.INFO,
                        )

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getServerUrl(self) -> Optional[str]:
        return self.getOrDefault(self.server_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getResource(self) -> str:
        return self.getOrDefault(self.resource)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilterByResource(self) -> Optional[str]:
        return self.getOrDefault(self.filter_by_resource)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilterParameter(self) -> Optional[str]:
        return self.getOrDefault(self.filter_parameter)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAdditionalParameters(self) -> Optional[List[str]]:
        return self.getOrDefault(self.additional_parameters)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getIdView(self) -> Optional[str]:
        return self.getOrDefault(self.id_view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(
        self,
    ) -> Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> Optional[int]:
        return self.getOrDefault(self.limit)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAction(self) -> Optional[str]:
        return self.getOrDefault(self.action)

    # noinspection PyPep8Naming
    def getActionPayload(self) -> Optional[Dict[str, Any]]:
        return self.getOrDefault(self.action_payload)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getIncludeOnlyProperties(self) -> Optional[List[str]]:
        return self.getOrDefault(self.include_only_properties)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getPageSize(self) -> Optional[int]:
        return self.getOrDefault(self.page_size)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getBatchSize(self) -> Optional[int]:
        return self.getOrDefault(self.batch_size)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLastUpdateAfter(self) -> Optional[datetime]:
        return self.getOrDefault(self.last_updated_after)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLastUpdatedBefore(self) -> Optional[datetime]:
        return self.getOrDefault(self.last_updated_before)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSortFields(self) -> Optional[List[SortField]]:
        return self.getOrDefault(self.sort_fields)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthServerUrl(self) -> Optional[str]:
        return self.getOrDefault(self.auth_server_url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthClientId(self) -> Optional[str]:
        return self.getOrDefault(self.auth_client_id)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthClientSecret(self) -> Optional[str]:
        return self.getOrDefault(self.auth_client_secret)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthLoginToken(self) -> Optional[str]:
        return self.getOrDefault(self.auth_login_token)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAuthScopes(self) -> Optional[List[str]]:
        return self.getOrDefault(self.auth_scopes)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSeparateBundleResources(self) -> bool:
        return self.getOrDefault(self.separate_bundle_resources)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getExpandFhirBundle(self) -> bool:
        return self.getOrDefault(self.expand_fhir_bundle)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getErrorOnResultCount(self) -> bool:
        return self.getOrDefault(self.error_on_result_count)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAcceptType(self) -> Optional[str]:
        return self.getOrDefault(self.accept_type)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getContentType(self) -> Optional[str]:
        return self.getOrDefault(self.content_type)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAdditionalRequestHeaders(self) -> Optional[Dict[str, str]]:
        return self.getOrDefault(self.additional_request_headers)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getAcceptEncoding(self) -> Optional[str]:
        return self.getOrDefault(self.accept_encoding)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMode(self) -> str:
        return self.getOrDefault(self.mode)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getIgnoreStatusCodes(self) -> Optional[List[int]]:
        return self.getOrDefault(self.ignore_status_codes)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getVerifyCountsMatch(self) -> Optional[bool]:
        return self.getOrDefault(self.verify_counts_match)



import json
from typing import (
    List,
    Iterable,
    Optional,
    Dict,
    Any,
    Tuple,
    Union,
    AsyncGenerator,
)

# noinspection PyPackageRequirements
from aiohttp import ClientResponse

# noinspection PyPackageRequirements
from pyspark import SparkFiles

from spark_pipeline_framework.register import register
from spark_pipeline_framework.transformers.http_data_receiver.v5.common import (
    RESPONSE_PROCESSOR_TYPE,
)
from spark_pipeline_framework.transformers.http_data_receiver.v5.http_data_receiver_parameters import (
    HttpDataReceiverParameters,
)
from spark_pipeline_framework.utilities.api_helper.v2.http_request import (
    HelixHttpRequest,
    RequestType,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_batch_function_run_context import (
    AsyncPandasBatchFunctionRunContext,
)
from spark_pipeline_framework.utilities.oauth2_helpers.v3.oauth2_client_credentials_flow import (
    OAuth2ClientCredentialsFlow,
    OAuth2Credentails,
)


class HttpDataReceiverProcessor:

    # noinspection PyUnusedLocal
    @staticmethod
    async def send_chunk_request(
        run_context: AsyncPandasBatchFunctionRunContext,
        rows: Iterable[Dict[str, Any]],
        parameters: HttpDataReceiverParameters,
        additional_parameters: Optional[Dict[str, Any]],
    ) -> AsyncGenerator[Dict[str, Any], None]:
        """
        This function processes a partition calling fhir server for each row in the partition

        This has to be a static function to avoid creating a closure around a class
        https://spark.apache.org/docs/latest/rdd-programming-guide.html#passing-functions-to-spark


        :param run_context: run context
        :param rows: rows to process
        :param parameters: FhirReceiverParameters
        :param additional_parameters: additional parameters
        :return: rows
        """
        register()
        row: List[Dict[str, Any]]
        async for row in HttpDataReceiverProcessor.process_rows_async(
            partition_index=run_context.partition_index,
            rows=rows,
            response_processor=parameters.response_processor,
            raise_error=parameters.raise_error,
            credentials=parameters.credentials,
            auth_url=parameters.auth_url,
            cert=parameters.cert,
            verify=parameters.verify,
        ):
            yield {
                "rows": row,
            }

    @staticmethod
    async def create_access_token_async(
        *, auth_url: str, credentials: Optional[OAuth2Credentails]
    ) -> str:
        """
        Factory function to create access token

        :param auth_url: OAuth token URL
        :param credentials: authentication credentials like client id/secrets
        """
        assert credentials
        oauth2_client_credentials_flow: OAuth2ClientCredentialsFlow = (
            OAuth2ClientCredentialsFlow(
                auth_url=auth_url,
                auth_credentials=credentials,
                progress_logger=None,
            )
        )

        access_token: Optional[str] = (
            await oauth2_client_credentials_flow.get_token_async()
        )
        assert access_token

        return access_token

    @staticmethod
    async def process_rows_async(
        partition_index: int,
        rows: Iterable[Dict[str, Any]],
        response_processor: RESPONSE_PROCESSOR_TYPE,
        raise_error: bool,
        credentials: Optional[OAuth2Credentails],
        auth_url: Optional[str],
        cert: Optional[Union[str, Tuple[str, str]]] = None,
        verify: Optional[Union[bool, str]] = None,
    ) -> AsyncGenerator[List[Dict[str, Any]], None]:
        headers = {}

        if credentials or auth_url:
            assert credentials
            assert auth_url
            auth_token: str = await HttpDataReceiverProcessor.create_access_token_async(
                auth_url=auth_url, credentials=credentials
            )
            headers.update({"Authorization": f"Bearer {auth_token}"})

        # Assumes certs are distributed to the executors beforehand via SparkContext.addFile
        cert_files: Optional[Union[str, Tuple[str, str]]] = None
        if isinstance(cert, tuple):
            cert_files = SparkFiles.get(cert[0]), SparkFiles.get(cert[1])
        elif cert:
            cert_files = SparkFiles.get(cert)

        for row in rows:
            async for r in HttpDataReceiverProcessor.process_record_async(
                partition_index=partition_index,
                auth_url=auth_url,
                cert_files=cert_files,
                credentials=credentials,
                headers=headers,
                raise_error=raise_error,
                response_processor=response_processor,
                row=row,
                verify=verify,
            ):
                yield r

    # noinspection PyUnusedLocal
    @staticmethod
    async def process_record_async(
        *,
        partition_index: int,
        headers: Dict[str, Any],
        row: Dict[str, Any],
        response_processor: RESPONSE_PROCESSOR_TYPE,
        raise_error: bool,
        credentials: Optional[OAuth2Credentails],
        auth_url: Optional[str],
        cert_files: Optional[Union[str, Tuple[str, str]]] = None,
        verify: Optional[Union[bool, str]] = None,
    ) -> AsyncGenerator[List[Dict[str, Any]], None]:
        response: ClientResponse = await HttpDataReceiverProcessor.process_row_async(
            row=row,
            raise_error=False,  # We don't want to raise error in case of access token expiry
            base_headers=headers,
            cert=cert_files,
            verify=verify,
        )
        if auth_url and response.status == 401:
            access_token: str = (
                await HttpDataReceiverProcessor.create_access_token_async(
                    auth_url=auth_url, credentials=credentials
                )
            )
            headers.update({"Authorization": f"Bearer {access_token}"})
            response = await HttpDataReceiverProcessor.process_row_async(
                row=row,
                raise_error=raise_error,
                base_headers=headers,
                cert=cert_files,
                verify=verify,
            )
        elif raise_error:
            response.raise_for_status()
        data, is_error = await response_processor(response, json.loads(row["state"]))
        if not isinstance(data, list):
            data = [data]

        yield [
            dict(
                headers=headers,
                url=row["url"],
                status=response.status,
                is_error=is_error,
                error_data=json.dumps(item if is_error else None),
                success_data=json.dumps(None if is_error else item),
                state=json.loads(row["state"]),
            )
            for item in data
        ]

    @staticmethod
    async def process_row_async(
        row: Dict[str, Any],
        raise_error: bool,
        base_headers: Dict[str, Any],
        cert: Optional[Union[str, Tuple[str, str]]] = None,
        verify: Optional[Union[bool, str]] = None,
    ) -> ClientResponse:
        headers = json.loads(row["headers"])
        headers.update(base_headers)
        http_request = HelixHttpRequest(
            request_type=RequestType.GET,
            url=row["url"],
            headers=headers,
            raise_error=raise_error,
            cert=cert,
            verify=verify,
        )
        return await http_request.get_response_async()



import dataclasses
from typing import List, Dict, Any, AsyncGenerator, Optional

from pyspark.sql.types import (
    StringType,
    StructType,
)
from pyspark.sql.types import StructField, ArrayType

from spark_pipeline_framework.transformers.fhir_receiver.v2.structures.get_batch_error import (
    GetBatchError,
)


@dataclasses.dataclass
class GetBatchResult:
    resources: List[str]
    errors: List[GetBatchError]

    @staticmethod
    def get_schema() -> StructType:
        return StructType(
            [
                StructField("resources", ArrayType(StringType()), True),
                StructField("errors", ArrayType(GetBatchError.get_schema()), True),
            ]
        )

    def to_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    def append(self, result: "GetBatchResult") -> "GetBatchResult":
        self.resources = self.resources + result.resources
        self.errors = self.errors + result.errors
        return self

    @classmethod
    def from_list(cls, data: List["GetBatchResult"]) -> "GetBatchResult":
        return cls(
            resources=[f for r in data for f in r.resources],
            errors=[f for r in data for f in r.errors],
        )

    @staticmethod
    async def from_async_generator(
        generator: AsyncGenerator["GetBatchResult", None],
    ) -> Optional["GetBatchResult"]:
        """
        Reads a generator of FhirGetResponse and returns a single FhirGetResponse by appending all the FhirGetResponse

        :param generator: generator of FhirGetResponse items
        :return: FhirGetResponse
        """
        result: GetBatchResult | None = None
        async for value in generator:
            if not result:
                result = value
            else:
                result.append(value)

        assert result
        return result



from typing import Dict, Any, Optional

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import explode
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkStructToColumns(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view: str,
        output_view: str,
        column_to_explode: str,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        self.view: Param = Param(self, "view", "")  # type: ignore
        self._setDefault(view=view)

        self.output_view: Param = Param(self, "output_view", "")  # type: ignore
        self._setDefault(output_view=output_view)

        self.column_to_explode: Param = Param(self, "column_to_explode", "")  # type: ignore
        self._setDefault(column_to_explode=column_to_explode)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        output_view: str = self.getOutputView()
        column_to_explode: str = self.getColumnToExplode()

        source_df = df.sparkSession.table(view)
        output_df = source_df.select(
            explode(column_to_explode).alias(column_to_explode)
        )
        output_df.createOrReplaceTempView(output_view)

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getOutputView(self) -> str:
        return self.getOrDefault(self.output_view)  # type: ignore

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getColumnToExplode(self) -> str:
        return self.getOrDefault(self.column_to_explode)  # type: ignore



import os
from os import path, makedirs
from pathlib import Path
from shutil import rmtree

from spark_pipeline_framework.utilities.file_downloader.file_downloader import (
    FileDownloader,
)


def test_file_downloader() -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")
    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)
    makedirs(temp_folder.joinpath("data"))

    file_url: str = (
        "https://files.pythonhosted.org/packages/47/6a/62e288da7bcda82b935ff0c6cfe542970f04e29c756b0e147251b2fb251f/wget-3.2.zip"
    )
    download_path: str = f"file://{os.path.join(temp_folder, 'data')}"

    # Act
    downloader = FileDownloader(
        url=file_url,
        download_path=download_path,
    )
    filename = downloader.download_files_from_url()

    # Assert
    local_file_protocol_length = len("file://")
    if download_path.split(":")[0] in ["file"]:
        assert os.path.exists(os.path.join(download_path[local_file_protocol_length:], filename))  # type: ignore



import time

import pytest

from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.base import Transformer
from pyspark.ml.feature import SQLTransformer
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.session import SparkSession
from pyspark.sql.types import StructType, StructField, StringType, IntegerType

from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.utilities.parallel_pipeline_executor.v1.parallel_pipeline_executor import (
    ParallelPipelineExecutor,
)


@pytest.mark.asyncio
async def test_can_run_pipelines_in_parallel(spark_session: SparkSession) -> None:
    my_schema = StructType(
        [
            StructField("Name", StringType(), False),
            StructField("Age", IntegerType(), False),
        ]
    )

    test_list = [
        ("Mike", 19),
        ("June", 18),
        ("Rachel", 16),
        ("Rob", 18),
        ("Scott", 17),
        ("Mike", 21),
        ("Mike", 15),
    ]

    df = spark_session.createDataFrame(test_list, schema=my_schema)

    executor = ParallelPipelineExecutor(max_tasks=2, progress_logger=ProgressLogger())
    executor.append(
        "Name",
        [
            SQLTransformer(statement="SELECT Name FROM __THIS__"),
            DummyDelayTransformer(delay=10),
        ],
    )
    executor.append(
        "age",
        [
            SQLTransformer(statement="SELECT Age FROM __THIS__"),
            DummyDelayTransformer(delay=10),
        ],
    )
    executor.append(
        "Name+Age",
        [
            SQLTransformer(statement="SELECT Name, Age FROM __THIS__"),
            DummyDelayTransformer(delay=5),
        ],
    )

    expected_count = 3

    actual_count = 0
    async for name, result_df in executor.transform_async(df, spark_session):
        actual_count += 1
        assert type(result_df) == DataFrame
        result_df.show()

    assert actual_count == expected_count


class DummyDelayTransformer(FrameworkTransformer):
    """
    dummy
    """

    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(self, delay: int) -> None:
        super().__init__()
        self.delay: Param[int] = Param(self, "delay", "")
        self._setDefault(delay=delay)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    # noinspection PyMethodMayBeStatic
    def _transform(self, df: DataFrame) -> DataFrame:
        time.sleep(float(self.getDelay()))
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def setDelay(self, value: int) -> Transformer:
        self._paramMap[self.delay] = value
        return self

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDelay(self) -> int:
        return self.getOrDefault(self.delay)



from typing import Dict, Any, Optional

import requests

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from requests import Response
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class HttpDataReceiver(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        url: str,
        view: str,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ) -> None:
        """
        Transformer to call and receive data from an API


        :param url: URL to request data from
        :param view: name of the view to read the response into
        :param name: name of transformer
        :param parameters: parameters
        :param progress_logger: progress logger
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.url: Param[str] = Param(self, "url", "")
        self._setDefault(url=None)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=None)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        url: str = self.getUrl()
        view = self.getView()
        name: Optional[str] = self.getName()

        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        with ProgressLogMetric(
            name=f"{name}_http_data_receiver", progress_logger=progress_logger
        ):
            response: Response = requests.get(
                url, headers={"User-Agent": "helix/1.0.0"}
            )

            if response.ok:
                self.logger.info(f"Successfully retrieved: {url}")

                text = response.text
                self.logger.info(text)
                df2 = df.sparkSession.read.json(
                    df.sparkSession.sparkContext.parallelize([text])
                )

                df2.createOrReplaceTempView(view)
            else:
                self.logger.error(f"Receive failed [{response.status_code}]: {url} ")
                error_text: str = response.text
                self.logger.error(error_text)
                raise Exception(error_text)

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getUrl(self) -> str:
        return self.getOrDefault(self.url)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)



from typing import Optional
from datetime import datetime, timezone, timedelta

from spark_pipeline_framework.event_loggers.event_logger import EventLogger
from spark_pipeline_framework.utilities.slack.base_slack_client import BaseSlackClient

from spark_pipeline_framework.utilities.slack.slack_client import SlackClient
from spark_pipeline_framework.utilities.slack.slack_client_native import (
    SlackClientNative,
)


class SlackEventLogger(EventLogger):
    def __init__(
        self,
        id_: str,
        slack_token: str,
        slack_channel: str,
        bot_user_name: str,
        slack_error_channel: Optional[str] = None,
        log_placeholder_url: Optional[str] = None,
        flow_run_name: Optional[str] = None,
        team_name: Optional[str] = None,
        backoff: bool = True,
        use_native_slack_client: bool = False,
    ) -> None:
        """
        Logs events to Slack

        :param id_: unique id for this logger
        :param slack_token: token to auth with Slack
        :param slack_channel: channel to post message into
        :param slack_error_channel: channel to post error message into
        :param bot_user_name: user name to use when posting messages
        :param log_placeholder_url: url to logs with a placeholder for flow_run_name to be replaced by id
                e.g., https://app.groundcover.com/logs?backendId=groundcover&filters=%255B%2522k8s.pod.label.flow_run_name%253A{flow_run_name}%2522%255D&src_cluster=spark-prod-ue1&start={from_time}&end={to_time}
        :param backoff: whether to backoff sending too many messages to slack
        :param flow_run_name: flow run name to set in log_placeholder_url
        :param use_native_slack_client: whether to use the native slack client or use plain HTTP post
        """
        super().__init__()
        assert slack_token
        assert slack_channel
        assert bot_user_name
        self.id_: str = id_
        self.team_name = team_name
        self.slack_token: str = slack_token
        self.slack_channel: str = slack_channel
        self.slack_error_channel: Optional[str] = slack_error_channel
        self.slack_user_name: str = bot_user_name
        self.backoff: bool = backoff
        self.slack_client: BaseSlackClient = (
            SlackClientNative(
                slack_token=slack_token,
                channel=slack_channel,
                bot_user_name=bot_user_name,
            )
            if use_native_slack_client
            else SlackClient(
                slack_token=slack_token,
                channel=slack_channel,
                bot_user_name=bot_user_name,
            )
        )
        self.slack_error_client: Optional[BaseSlackClient] = (
            (
                SlackClientNative(
                    slack_token=slack_token,
                    channel=slack_error_channel,
                    bot_user_name=bot_user_name,
                )
                if use_native_slack_client
                else SlackClient(
                    slack_token=slack_token,
                    channel=slack_error_channel,
                    bot_user_name=bot_user_name,
                )
            )
            if slack_error_channel
            else None
        )
        self.flow_run_name: Optional[str] = flow_run_name
        self.log_placeholder_url: Optional[str] = log_placeholder_url
        self.has_sent_log_url: bool = False

    def log_progress_event(
        self,
        event_name: str,
        current: int,
        total: int,
        event_format_string: str,
        backoff: bool = True,
    ) -> None:
        # for first 10 batches send every time
        # for 10 to 100 send every 10 batches
        # after 100 send every 100 batches
        if (
            not self.backoff
            or not backoff
            or (current < 10)
            or (current < 100 and current % 10 == 0)
            or (current < 1000 and current % 100 == 0)
            or (current < 10000 and current % 1000 == 0)
            or (current < 100000 and current % 10000 == 0)
            or (current < 1000000 and current % 100000 == 0)
        ):
            self.slack_client.post_message_to_slack(
                event_format_string.format(event_name, current, total)
            )

    def log_event(self, event_name: str, event_text: str) -> None:
        self.slack_client.post_message_to_slack(event_text)
        if not self.has_sent_log_url:
            if self.log_placeholder_url:
                log_url = self.get_log_url()
                if log_url:
                    self.slack_client.post_message_to_slack(log_url)
            self.has_sent_log_url = True

    def log_exception(self, event_name: str, event_text: str, ex: Exception) -> None:
        # don't send full exception to slack since it can have PHI
        message = (
            f"*Event Name:* Helix Pipeline Failure\n"
            f"{f'*Team Owner:* {self.team_name}\n' if self.team_name else ''}"
            f"*Flow Run Name:* {self.flow_run_name}\n"
            f"*Deployment Name:* {event_name}\n"
            f"*Run ID:* {self.id_}\n"
            f"*Logs URL:* <{self.get_log_url()}|View Logs>"
        )
        response = self.slack_client.post_message_to_slack(message)
        thread_ts = response.data.get("ts") if response else None

        if thread_ts:
            self.slack_client.post_message_to_slack(
                f"*Event Details:* {event_text}\n*Error Details:* ```{str(ex)}```",
                use_conversation_threads=True,
                slack_thread=thread_ts,
            )

        if self.slack_error_client:
            response = self.slack_error_client.post_message_to_slack(message)
            thread_ts = response.data.get("ts") if response else None

            if thread_ts:
                self.slack_error_client.post_message_to_slack(
                    f"Error Details: {str(ex)}",
                    use_conversation_threads=True,
                    slack_thread=thread_ts,
                )

    def get_log_url(self) -> Optional[str]:
        if not self.log_placeholder_url:
            return None

        current_time = (
            (datetime.now(timezone.utc) + timedelta(days=1))
            .isoformat(timespec="milliseconds")
            .replace("+00:00", "Z")
        )
        from_time = (
            (datetime.now(timezone.utc) - timedelta(days=1))
            .isoformat(timespec="milliseconds")
            .replace("+00:00", "Z")
        )

        return self.log_placeholder_url.format(
            flow_run_name=f"{self.flow_run_name}",
            from_time=from_time,
            to_time=current_time,
        )



from typing import Dict, Any, Callable, Union, List

from spark_auto_mapper.automappers.automapper_base import AutoMapperBase

from library.features.carriers.v1.features_carriers_v1 import FeaturesCarriersV1
from library.features.carriers_python.v1.features_carriers_python_v1 import (
    FeaturesCarriersPythonV1,
)
from spark_pipeline_framework.pipelines.v2.framework_pipeline import FrameworkPipeline
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.proxy_generator.python_transformer_helpers import (
    get_python_function_from_location,
)
from spark_pipeline_framework.transformers.framework_csv_loader.v1.framework_csv_loader import (
    FrameworkCsvLoader,
)
from spark_pipeline_framework.transformers.framework_drop_views_transformer.v1.framework_drop_views_transformer import (
    FrameworkDropViewsTransformer,
)
from spark_pipeline_framework.transformers.framework_if_else_transformer.v1.framework_if_else_transformer import (
    FrameworkIfElseTransformer,
)
from spark_pipeline_framework.transformers.framework_json_exporter.v1.framework_json_exporter import (
    FrameworkJsonExporter,
)
from spark_pipeline_framework.transformers.framework_loop_transformer.v1.framework_loop_transformer import (
    FrameworkLoopTransformer,
)
from spark_pipeline_framework.transformers.framework_mapping_runner.v1.framework_mapping_runner import (
    FrameworkMappingLoader,
)


class LoopingPipeline(FrameworkPipeline):
    def __init__(
        self,
        parameters: Dict[str, Any],
        progress_logger: ProgressLogger,
        max_number_of_runs: int = 1,
    ):
        super(LoopingPipeline, self).__init__(
            parameters=parameters,
            progress_logger=progress_logger,
            run_id="87654321",
            client_name="client_foo",
            vendor_name="vendor_foo",
        )

        mapping_function: Callable[
            [Dict[str, Any]], Union[AutoMapperBase, List[AutoMapperBase]]
        ] = get_python_function_from_location(
            location=str(parameters["feature_path"]),
            import_module_name=".mapping",
            function_name="mapping",
        )

        self.steps = [
            FrameworkLoopTransformer(
                name="FrameworkLoopTransformer",
                parameters=parameters,
                progress_logger=progress_logger,
                sleep_interval_in_seconds=2,
                max_number_of_runs=max_number_of_runs,
                stages=[
                    FrameworkDropViewsTransformer(
                        name="",
                        parameters=parameters,
                        progress_logger=progress_logger,
                        views=["foo"],
                    ),
                    FrameworkCsvLoader(
                        name="FrameworkCsvLoader",
                        view="flights",
                        file_path=parameters["flights_path"],
                        parameters=parameters,
                        progress_logger=progress_logger,
                    ),
                    FeaturesCarriersV1(
                        parameters=parameters, progress_logger=progress_logger
                    ),
                    FrameworkIfElseTransformer(
                        enable=True,
                        progress_logger=progress_logger,
                        stages=[
                            FeaturesCarriersPythonV1(
                                parameters=parameters, progress_logger=progress_logger
                            ),
                        ],
                    ),
                    FrameworkMappingLoader(
                        view="members",
                        mapping_function=mapping_function,
                        parameters=parameters,
                        progress_logger=progress_logger,
                    ),
                    FrameworkJsonExporter(
                        file_path=parameters["export_path"],
                        view="flights",
                        name="FrameworkJsonExporter",
                        parameters=parameters,
                        progress_logger=progress_logger,
                    ),
                ],
            )
        ]



import dataclasses
from datetime import datetime
from logging import Logger
from typing import Any, Dict, List, Optional, AsyncGenerator

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_batch_function_run_context import (
    AsyncPandasBatchFunctionRunContext,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.address_standardizer import (
    AddressStandardizer,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.cache.cache_handler import (
    CacheHandler,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.base_vendor_api_response import (
    BaseVendorApiResponse,
)
from spark_pipeline_framework.utilities.spark_partition_information.v1.spark_partition_information import (
    SparkPartitionInformation,
)


@dataclasses.dataclass
class AddressStandardizationParameters:
    total_partitions: int
    address_column_mapping: Dict[str, str]
    standardizing_vendor: StandardizingVendor[BaseVendorApiResponse]
    cache_handler: CacheHandler
    geolocation_column_prefix: Optional[str]
    # add parameters to pass to async function below here
    log_level: str = "INFO"


class AddressStandardizationProcessor:
    # noinspection PyUnusedLocal
    @staticmethod
    async def standardize_list(
        run_context: AsyncPandasBatchFunctionRunContext,
        input_values: List[Dict[str, Any]],
        parameters: Optional[AddressStandardizationParameters],
        additional_parameters: Optional[Dict[str, Any]],
    ) -> AsyncGenerator[Dict[str, Any], None]:
        """
        Standardize a list of raw addresses.  raw address is a dictionary with the following keys
        address1, address2, city, state, zip
        Returns a list of dictionary raw_addresses with the following keys
        address1, address2, city, state, zip, latitude, longitude

        :param input_values:
        :param run_context: the run context
        :param parameters:
        :param additional_parameters:
        :return:
        """
        assert parameters, "parameters is required"
        address_column_mapping: Dict[str, str] = parameters.address_column_mapping
        assert address_column_mapping, "address_column_mapping is required"
        cache_handler: CacheHandler = parameters.cache_handler
        assert cache_handler, "cache_handler is required"
        standardizing_vendor: StandardizingVendor[BaseVendorApiResponse] = (
            parameters.standardizing_vendor
        )
        assert standardizing_vendor, "standardizing_vendor is required"

        geolocation_column_prefix: Optional[str] = parameters.geolocation_column_prefix

        assert all(r.get(address_column_mapping["address_id"]) for r in input_values)
        logger: Logger = get_logger(
            __name__,
            level=(
                parameters.log_level if parameters and parameters.log_level else "INFO"
            ),
        )
        spark_partition_information: SparkPartitionInformation = (
            SparkPartitionInformation.from_current_task_context(
                chunk_index=run_context.chunk_index,
            )
        )
        message: str = f"FhirReceiverProcessor.process_partition"
        # Format the time to include hours, minutes, seconds, and milliseconds
        formatted_time = datetime.now().strftime("%H:%M:%S.%f")[:-3]
        formatted_message: str = (
            f"{formatted_time}: "
            f"{message}"
            f" | Partition: {run_context.partition_index}"
            f" | Chunk: {run_context.chunk_index}"
            f" | range: {run_context.chunk_input_range.start}-{run_context.chunk_input_range.stop}"
            f" | {spark_partition_information}"
        )
        logger.info(formatted_message)

        try:
            # create raw address raw_addresses to send to the standardization module...
            raw_address_list: List[RawAddress] = [
                RawAddress(
                    address_id=raw_address[address_column_mapping["address_id"]],
                    line1=raw_address[address_column_mapping["line1"]],
                    line2=raw_address[address_column_mapping["line2"]],
                    city=raw_address[address_column_mapping["city"]],
                    state=raw_address[address_column_mapping["state"]],
                    zipcode=raw_address[address_column_mapping["zipcode"]],
                )
                for raw_address in input_values
            ]
            # standardize the raw addresses which also calculates the lat/long
            standard_addresses: List[
                StandardizedAddress
            ] = await AddressStandardizer().standardize_async(
                raw_addresses=raw_address_list,
                cache_handler_obj=cache_handler,
                vendor_obj=standardizing_vendor,
            )
            logger.debug(
                f"Received result"
                f" | Partition: {run_context.partition_index}"
                f" | Chunk: {run_context.chunk_index}"
                f" | Vendor: {standardizing_vendor.get_vendor_name()}"
                f" | Count: {len(standard_addresses)}"
            )
            assert len(standard_addresses) == len(
                input_values
            ), f"Length of output != Length of input: {len(standard_addresses)} != {len(input_values)}"
            # map standard address back to a list of dictionary raw_addresses
            standard_address_list: List[Dict[str, Optional[str]]] = [
                {
                    f"{geolocation_column_prefix}latitude": standard_address.latitude,
                    f"{geolocation_column_prefix}longitude": standard_address.longitude,
                    address_column_mapping["line1"]: standard_address.line1,
                    address_column_mapping["line2"]: standard_address.line2,
                    address_column_mapping["city"]: standard_address.city,
                    address_column_mapping["state"]: standard_address.state,
                    address_column_mapping["zipcode"]: standard_address.zipcode,
                }
                for standard_address in standard_addresses
            ]
            for standard_address in standard_address_list:
                yield standard_address
        except Exception as e1:
            raise e1



from pyspark.sql import SparkSession, DataFrame
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.athena_table_creator.v1.athena_table_creator import (
    AthenaTableCreator,
)
from spark_pipeline_framework.utilities.athena.athena import Athena
from spark_pipeline_framework.utilities.athena.athena_source_file_type import (
    AthenaSourceFileType,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)

from unittest import mock
from unittest.mock import MagicMock


@mock.patch.object(Athena, "drop_create_table")
def test_athena_table_creator(
    athena_mock: MagicMock, spark_session: SparkSession
) -> None:
    # Arrange
    df: DataFrame = create_empty_dataframe(spark_session=spark_session)
    df.createTempView("view_name")

    # Act
    with ProgressLogger() as progress_logger:
        AthenaTableCreator(
            view="view_name",
            schema_name="schema",
            table_name="table",
            s3_source_path="source_path",
            s3_temp_folder="s3://bwell-helix-log-scratchpad-ue1/athena_query_results/",
            source_file_type=AthenaSourceFileType.PARQUET,
            progress_logger=progress_logger,
        ).transform(df)

    # Assert
    athena_mock.assert_called_with(
        "table",
        "schema",
        mock.ANY,
        "source_path",
        AthenaSourceFileType.PARQUET,
        s3_temp_folder="s3://bwell-helix-log-scratchpad-ue1/athena_query_results/",
    )



from pyspark.sql.types import (
    StructType,
    StructField,
    IntegerType,
    StringType,
    ArrayType,
)


class FhirMergeResponseSchema:
    """
    This class provides names for columns in FhirMergeResponse

    Should match to https://github.com/icanbwell/helix.fhir.client.sdk/blob/main/helix_fhir_client_sdk/responses/fhir_merge_response.py
    """

    request_id = "request_id"
    url = "url"
    responses = "responses"
    error = "error"
    access_token = "access_token"
    status = "status"
    data = "data"

    @staticmethod
    def get_schema() -> StructType:
        """
        Returns the schema of FhirMergeResponse

        Should match to https://github.com/icanbwell/helix.fhir.client.sdk/blob/main/helix_fhir_client_sdk/responses/fhir_merge_response.py
        """
        response_schema = StructType(
            [
                StructField("request_id", StringType(), nullable=True),
                StructField("url", StringType(), nullable=False),
                StructField("responses", ArrayType(StringType()), nullable=False),
                StructField("error", StringType(), nullable=True),
                StructField("access_token", StringType(), nullable=True),
                StructField("status", IntegerType(), nullable=False),
                StructField("data", StringType(), nullable=False),
            ]
        )
        return response_schema



import json
import math
from typing import Any, Dict, List, Optional, cast, Union, Tuple

from more_itertools import chunked
from pyspark import StorageLevel
from pyspark.sql.types import Row
from pyspark.rdd import RDD
from pyspark.sql.functions import col, from_json
from pyspark.sql.types import (
    StructField,
    StringType,
    StructType,
    MapType,
    IntegerType,
    BooleanType,
)

from spark_pipeline_framework.transformers.http_data_receiver.v4.http_data_receiver_processor import (
    HttpDataReceiverProcessor,
    RESPONSE_PROCESSOR_TYPE,
    REQUEST_GENERATOR_TYPE,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.oauth2_helpers.v2.oauth2_client_credentials_flow import (
    OAuth2Credentails,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


class HttpDataReceiver(FrameworkTransformer):
    """
    This is a generic class to call a http api and return the response
    """

    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        name: str,
        success_view: str,
        error_view: str,
        http_request_generator: REQUEST_GENERATOR_TYPE,
        response_processor: RESPONSE_PROCESSOR_TYPE,
        success_schema: Optional[StructType] = None,
        error_schema: Optional[StructType] = None,
        num_partition: Optional[int] = None,
        batch_size: int = 1000,
        items_per_partition: Optional[int] = None,
        cache_storage_level: Optional[StorageLevel] = None,
        credentials: Optional[OAuth2Credentails] = None,
        auth_url: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        run_sync: bool = False,
        raise_error: bool = False,
        progress_logger: Optional[ProgressLogger] = None,
        cert: Optional[Union[str, Tuple[str, str]]] = None,
        verify: Optional[Union[bool, str]] = None,
    ) -> None:
        """
        Transformer to call and receive data from an API

        :param name: name of transformer
        :param success_view: name of the view to read the response into
        :param error_view: (Optional) log the details of the api failure into `error_view` view.
        :param http_request_generator: Generator to build next http request
        :param response_processor: it can change the result before loading to spark df
        :param success_schema: Schema for success response
        :param error_schema: Schema for error response
        :param num_partition: Number of batches
        :param batch_size: Size of a partition, used in internal processing like converting requests to view
        :param items_per_partition: Number of items to process per partition
        :param cache_storage_level: (Optional) how to store the cache:
                                    https://sparkbyexamples.com/spark/spark-dataframe-cache-and-persist-explained/.
        :param credentials: OAuth2 credentails
        :param auth_url: OAuth2 token URL
        :param parameters: parameters
        :param run_sync: process the items linearly
        :param raise_error: (Optional) Raise error in case of api failure
        :param progress_logger: progress logger
        :param cert: certificate or ca bundle file path
        :param verify: controls whether the SSL certificate of the server should be verified when making HTTPS requests.
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.name: Param[str] = Param(self, "name", "")
        self._setDefault(name=None)

        self.success_view: Param[str] = Param(self, "success_view", "")
        self._setDefault(success_view=None)

        self.error_view: Param[str] = Param(self, "error_view", "")
        self._setDefault(error_view=None)

        self.http_request_generator: Param[REQUEST_GENERATOR_TYPE] = Param(
            self, "http_request_generator", ""
        )
        self._setDefault(http_request_generator=None)

        self.response_processor: Param[RESPONSE_PROCESSOR_TYPE] = Param(
            self, "response_processor", ""
        )
        self._setDefault(response_processor=None)

        self.success_schema: Param[Optional[StructType]] = Param(
            self, "success_schema", ""
        )
        self._setDefault(success_schema=success_schema)

        self.error_schema: Param[Optional[StructType]] = Param(self, "error_schema", "")
        self._setDefault(error_schema=error_schema)

        self.num_partition: Param[Optional[int]] = Param(self, "num_partition", "")
        self._setDefault(num_partition=None)

        self.batch_size: Param[int] = Param(self, "batch_size", "")
        self._setDefault(batch_size=batch_size)

        self.items_per_partition: Param[Optional[int]] = Param(
            self, "items_per_partition", ""
        )
        self._setDefault(items_per_partition=None)

        self.cache_storage_level: Param[Optional[StorageLevel]] = Param(
            self, "cache_storage_level", ""
        )
        self._setDefault(cache_storage_level=None)

        self.credentials: Param[Optional[OAuth2Credentails]] = Param(
            self, "credentials", ""
        )
        self._setDefault(credentials=None)

        self.auth_url: Param[Optional[str]] = Param(self, "auth_url", "")
        self._setDefault(auth_url=None)

        self.run_sync: Param[bool] = Param(self, "run_sync", "")
        self._setDefault(run_sync=run_sync)

        self.raise_error: Param[bool] = Param(self, "raise_error", "")
        self._setDefault(raise_error=raise_error)

        self.cert: Param[Optional[Union[str, Tuple[str, str]]]] = Param(
            self, "cert", ""
        )
        self._setDefault(cert=cert)

        self.verify: Param[Optional[Union[bool, str]]] = Param(self, "verify", "")
        self._setDefault(verify=verify)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        # Setting the variables
        name: str = self.getOrDefault(self.name)
        success_view: str = self.getOrDefault(self.success_view)
        error_view: str = self.getOrDefault(self.error_view)
        success_schema: Optional[StructType] = self.getOrDefault(self.success_schema)
        error_schema: Optional[StructType] = self.getOrDefault(self.error_schema)
        num_partition: Optional[int] = self.getOrDefault(self.num_partition)
        batch_size: int = self.getOrDefault(self.batch_size)
        items_per_partition: Optional[int] = self.getOrDefault(self.items_per_partition)
        http_request_generator: REQUEST_GENERATOR_TYPE = self.getOrDefault(
            self.http_request_generator
        )
        response_processor: RESPONSE_PROCESSOR_TYPE = self.getOrDefault(
            self.response_processor
        )
        cache_storage_level: Optional[StorageLevel] = self.getOrDefault(
            self.cache_storage_level
        )
        credentials: Optional[OAuth2Credentails] = self.getOrDefault(self.credentials)
        auth_url: Optional[str] = self.getOrDefault(self.auth_url)
        run_sync: bool = self.getOrDefault(self.run_sync)
        raise_error: bool = self.getOrDefault(self.raise_error)
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        cert: Optional[Union[str, Tuple[str, str]]] = self.getOrDefault(self.cert)
        verify: Optional[Union[bool, str]] = self.getOrDefault(self.verify)

        with ProgressLogMetric(
            name=f"{name}_http_data_receiver_v4", progress_logger=progress_logger
        ):
            requests_df: DataFrame = create_empty_dataframe(
                df.sparkSession,
                StructType(
                    [
                        StructField("url", StringType()),
                        StructField("headers", StringType()),
                        StructField("state", StringType()),
                    ]
                ),
            )
            for requests in chunked(
                http_request_generator(df, progress_logger), batch_size
            ):
                # Create the Dataframe
                view_data = [
                    [
                        request.url,
                        json.dumps(request.headers),
                        json.dumps(extra_context),
                    ]
                    for request, extra_context in requests
                ]
                df_ = df.sparkSession.createDataFrame(
                    view_data, ["url", "headers", "state"]
                )

                # Append the Dataframe
                requests_df = requests_df.union(df_)
                requests_df.createOrReplaceTempView("requests_view")

            desired_partitions: int = self.get_desired_partitions(
                num_partition=num_partition,
                items_per_partition=items_per_partition,
                df=requests_df,
            )

            row_schema = StructType(
                [
                    StructField("headers", MapType(StringType(), StringType())),
                    StructField("url", StringType()),
                    StructField("status", IntegerType()),
                    StructField("is_error", BooleanType()),
                    StructField("error_data", StringType()),
                    StructField("success_data", StringType()),
                    StructField("state", StringType()),
                ]
            )

            if run_sync:
                rows: List[Row] = requests_df.collect()
                result_rows = HttpDataReceiverProcessor.process_rows(
                    partition_index=0,
                    rows=rows,
                    response_processor=response_processor,
                    raise_error=raise_error,
                    credentials=credentials,
                    auth_url=auth_url,
                    cert=cert,
                    verify=verify,
                )

                # Create success view
                success = filter(lambda row: not row["is_error"], result_rows)
                df_success: DataFrame = (
                    df.sparkSession.createDataFrame(  # type:ignore[type-var]
                        [s.asDict(recursive=True) for s in success], schema=row_schema
                    )
                )
                json_schema = self.infer_schema_json_string_column(
                    df_success, "success_data"
                )
                self.copy_and_drop_column(
                    df_success, "success_data", "data", success_view, json_schema
                )

                # Create error view
                error = filter(lambda row: row["is_error"], result_rows)
                df_errors: DataFrame = (
                    df.sparkSession.createDataFrame(  # type:ignore[type-var]
                        [e.asDict(recursive=True) for e in error], schema=row_schema
                    )
                )
                json_schema = self.infer_schema_json_string_column(
                    df_errors, "error_data"
                )
                self.copy_and_drop_column(
                    df_errors, "error_data", "data", error_view, json_schema
                )
            else:
                rdd: RDD[Row] = requests_df.repartition(
                    desired_partitions
                ).rdd.mapPartitionsWithIndex(
                    lambda partition_index, rows_to_query: HttpDataReceiverProcessor.process_rows(
                        partition_index=partition_index,
                        rows=rows_to_query,
                        response_processor=response_processor,
                        raise_error=raise_error,
                        credentials=credentials,
                        auth_url=auth_url,
                        cert=cert,
                        verify=verify,
                    )
                )
                rdd = (
                    rdd.cache()
                    if cache_storage_level is None
                    else rdd.persist(storageLevel=cache_storage_level)
                )

                result_df: DataFrame = rdd.toDF(schema=row_schema)

                # Create success view
                df_success = result_df.where(result_df["is_error"] == False)
                json_schema = success_schema or self.infer_schema_json_string_column(
                    df_success, "success_data"
                )
                self.copy_and_drop_column(
                    df_success, "success_data", "data", success_view, json_schema
                )

                # Create error view
                df_errors = result_df.where(result_df["is_error"] == True)
                json_schema = error_schema or self.infer_schema_json_string_column(
                    df_errors, "error_data"
                )
                self.copy_and_drop_column(
                    df_errors, "error_data", "data", error_view, json_schema
                )

        return df

    def infer_schema_json_string_column(self, df: DataFrame, col_: str) -> StructType:
        """
        Infer json schema from `col_` column

        :param df: Dataframe to be processed.
        :param col_: Source column name
        """
        json_schema = df.sparkSession.read.json(
            df.rdd.map(lambda row: cast(str, row[col_]))
        ).schema
        return json_schema

    def copy_and_drop_column(
        self, df: DataFrame, col_: str, dest_col: str, view: str, schema: StructType
    ) -> None:
        """
        Copy the `col_` column to `dest_col` column with provided schema

        :param df: Dataframe to be processed.
        :param col_: source column
        :param dest_col: destination column
        :param view: Name of the view where the dataframe will be saved
        :param schema: schema of the `dest_col` column
        """
        df = df.withColumn(dest_col, from_json(col(col_), schema))
        df = df.drop("success_data", "error_data", "is_error")
        df.createOrReplaceTempView(view)

    def get_desired_partitions(
        self,
        *,
        df: DataFrame,
        num_partition: Optional[int] = None,
        items_per_partition: Optional[int] = None,
    ) -> int:
        """
        Get the desired partitions based on num_partition, items_per_partition and dataframe

        :param num_partition: number of desired partitions
        :param items_per_partition: number of items in a partitions
        :param df: Dataframe which will be divided into partitions
        """
        desired_partitions: int
        if num_partition:
            desired_partitions = num_partition
        else:
            row_count: int = df.count()
            desired_partitions = (
                math.ceil(row_count / items_per_partition)
                if items_per_partition and items_per_partition > 0
                else row_count
            ) or 1
        self.logger.info(f"Total Batches: {desired_partitions}")
        return desired_partitions



import json
from typing import Any, Dict, Iterable, List, Optional

from pyspark.sql.types import Row

from spark_pipeline_framework.register import register
from spark_pipeline_framework.utilities.api_helper.http_request import (
    HelixHttpRequest,
    RequestType,
)


class HttpDataSenderProcessor:
    @staticmethod
    # function that is called for each partition
    # noinspection PyUnusedLocal
    def send_partition_to_server(
        *,
        partition_index: int,
        rows: Iterable[Row],
        url: Optional[str],
        content_type: str,
        headers: Dict[str, Any],
        post_as_json_formatted_string: Optional[bool],
        parse_response_as_json: Optional[bool],
    ) -> Iterable[Row]:
        """
        This function processes a partition

        This has to be a static function to avoid creating a closure around a class
        https://spark.apache.org/docs/latest/rdd-programming-guide.html#passing-functions-to-spark


        """
        register()
        json_data_list: List[Dict[str, Any]] = [r.asDict(recursive=True) for r in rows]
        print(
            f"----- Sending batch {partition_index} containing {len(json_data_list)} rows -----"
        )

        # logger = get_logger(__name__)
        if len(json_data_list) == 0:
            yield Row(
                url=None,
                status=0,
                result=None,
                headers=None,
                payload=None,
                request_type=None,
            )

        assert url
        json_data: Dict[str, Any]
        for json_data in json_data_list:
            headers["Content-Type"] = content_type
            request: HelixHttpRequest = HelixHttpRequest(
                request_type=RequestType.POST,
                url=url,
                headers=headers,
                payload=json_data,
                post_as_json_formatted_string=post_as_json_formatted_string,
            )
            if parse_response_as_json:
                response_json = request.get_result()
                yield Row(
                    url=url,
                    status=response_json.status,
                    result=response_json.result,
                    headers=json.dumps(headers, default=str),
                    payload=json.dumps(json_data, default=str),
                    request_type=str(RequestType.POST),
                )
            else:
                response_text = request.get_text()
                yield Row(
                    url=url,
                    status=response_text.status,
                    result=response_text.result,
                    headers=json.dumps(headers, default=str),
                    payload=json.dumps(json_data, default=str),
                    request_type=str(RequestType.POST),
                )



import json
import math
from pathlib import Path
from typing import Any, Dict, Iterable, Iterator, List, Optional, Union

from spark_pipeline_framework.register import register
from spark_pipeline_framework.transformers.elasticsearch_sender.v1.elasticsearch_helpers import (
    send_json_bundle_to_elasticsearch,
)
from spark_pipeline_framework.transformers.elasticsearch_sender.v1.elasticsearch_result import (
    ElasticSearchResult,
)
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import to_json, struct, col
from pyspark.sql.types import Row
from pyspark.errors import AnalysisException
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    spark_is_data_frame_empty,
)


class ElasticSearchSender(FrameworkTransformer):
    @capture_parameters
    def __init__(
        self,
        index: str,
        file_path: Optional[Union[Path, str]] = None,
        view: Optional[str] = None,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        batch_size: int = 0,
        limit: int = -1,
        multi_line: bool = False,
        operation: str = "index",
        output_path: Optional[Union[Path, str]] = None,
    ):
        """
        Sends a folder or a view to an ElasticSearch server
        Either file_path or view must be set

        :param index: ElasticSearch index to update
        :param file_path: (Optional) read data from this folder to send
        :param view: (Optional) read data from this view
        :param batch_size: how many documents to process at one time
        :param multi_line: when reading data from folder, whether to expect multiline json files
        :param output_path: writes bulk output to this folder before sending to ES
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert file_path or view

        assert progress_logger

        self.logger = get_logger(__name__)

        self.view: Param[Optional[str]] = Param(self, "view", "")
        self._setDefault(view=view)

        self.file_path: Param[Optional[Union[Path, str]]] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.index: Param[str] = Param(self, "index", "")
        self._setDefault(index=index)

        self.limit: Param[int] = Param(self, "limit", "")
        self._setDefault(limit=limit)

        self.batch_size: Param[int] = Param(self, "batch_size", "")
        self._setDefault(batch_size=batch_size)

        self.multi_line: Param[bool] = Param(self, "multi_line", "")
        self._setDefault(multi_line=multi_line)

        self.operation: Param[str] = Param(self, "operation", "")
        self._setDefault(operation=operation)

        self.output_path: Param[Optional[Union[Path, str]]] = Param(
            self, "output_path", ""
        )
        self._setDefault(output_path=output_path)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: Optional[str] = self.getView()
        path: Optional[Union[Path, str]] = self.getFilePath()
        name: Optional[str] = self.getName()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        index: str = self.getIndex()
        batch_size: int = self.getBatchSize()
        limit: Optional[int] = self.getLimit()
        operation: str = self.getOperation()
        parameters: Optional[Dict[str, Any]] = self.getParameters()
        doc_id_prefix: Optional[str] = None
        if parameters is not None:
            doc_id_prefix = parameters.get("doc_id_prefix", None)

        # limit: int = self.getLimit()

        with ProgressLogMetric(
            name=f"{name or view}_elasticsearch_sender", progress_logger=progress_logger
        ):
            # read all the files at path into a dataframe
            path_to_files: str = str(path)
            try:
                if view:
                    json_df: DataFrame = df.sparkSession.table(view)
                    if "value" not in json_df.columns:
                        json_df = json_df.select(
                            to_json(struct(col("*"))).alias("value")
                        )
                else:
                    json_df = df.sparkSession.read.text(
                        path_to_files, pathGlobFilter="*.json", recursiveFileLookup=True
                    )
                if limit is not None and limit > 0:
                    json_df = json_df.limit(limit)
                # json_df.show(truncate=False, n=100000)
            except AnalysisException as e:
                if str(e).startswith("Path does not exist:"):
                    if progress_logger:
                        progress_logger.write_to_log(f"Folder {path_to_files} is empty")
                    return df
                raise

            row_count: int = json_df.count()
            if not spark_is_data_frame_empty(df=json_df):
                self.logger.info(
                    f"----- Sending {index} (rows={row_count}) to ElasticSearch server -----"
                )
                if not batch_size or batch_size < 1:
                    batch_size = 1
                desired_partitions: int = math.ceil(row_count / batch_size)
                self.logger.info(f"----- Total Batches: {desired_partitions}  -----")

                # function that is called for each partition
                def send_partition_to_server(
                    partition_index: int, rows: Iterable[Row]
                ) -> Iterable[Optional[ElasticSearchResult]]:
                    register()

                    json_data_list: List[str] = [r["value"] for r in rows]

                    if len(json_data_list) > 0:
                        self.logger.info(
                            f"Sending batch {partition_index}/{desired_partitions} "
                            f"containing {len(json_data_list)} rows "
                            f"to ES Server/{index}. [{name}].."
                        )
                        # send to server
                        response_json: ElasticSearchResult = (
                            send_json_bundle_to_elasticsearch(
                                json_data_list=json_data_list,
                                index=index,
                                operation=operation,
                                logger=self.logger,
                                doc_id_prefix=doc_id_prefix,
                            )
                        )
                        response_json.partition_index = partition_index
                        yield response_json
                    else:
                        self.logger.info(
                            f"Batch {partition_index}/{desired_partitions} is empty"
                        )
                        yield None

                response_from_server: Iterator[Optional[ElasticSearchResult]] = (
                    json_df.repartition(desired_partitions)
                    .rdd.mapPartitionsWithIndex(send_partition_to_server)
                    .toLocalIterator()
                )

                self.logger.info("---- Reply from server ----")
                response: Optional[ElasticSearchResult]
                for response in response_from_server:
                    if response:
                        self.logger.info(
                            f"partition {response.partition_index}  {response.url}. "
                            f"success={response.success}, failed={response.failed}"
                        )
                        if response.failed > 0:
                            self.logger.info(f"{json.dumps(response.payload)}")

                self.logger.info("---- End Reply from server ----")

        self.logger.info(
            f"----- Finished sending {index} (rows={row_count}) to ElasticSearch server  -----"
        )
        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> Optional[str]:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(self) -> Optional[Union[Path, str]]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getLimit(self) -> int:
        return self.getOrDefault(self.limit)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getIndex(self) -> str:
        return self.getOrDefault(self.index)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getBatchSize(self) -> int:
        return self.getOrDefault(self.batch_size)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getMultiline(self) -> int:
        return self.getOrDefault(self.multi_line)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> Optional[str]:
        return self.getOrDefault(self.name) or f"{self.getView()} - {self.getIndex()}"

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getOperation(self) -> str:
        return self.getOrDefault(self.operation)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getOutputPath(self) -> Optional[Union[Path, str]]:
        return self.getOrDefault(self.output_path)



from pyspark.sql import SparkSession
from unittest.mock import MagicMock, patch

from spark_pipeline_framework.transformers.framework_ddl_executor.v1.framework_ddl_executor import (
    FrameworkDDLExecutor,
)


def test_framework_databricks_ddl_executor_executes_ddl(
    spark_session: SparkSession,
) -> None:
    df = spark_session.createDataFrame([(1, "Alice")], ["id", "name"])
    ddl = """
    CREATE TABLE IF NOT EXISTS test_catalog.test_schema.test_table (
        id INT,
        name VARCHAR(255)
    )
    """
    jdbc_url = "jdbc:databricks://dummy-url"

    # Patch JVM and JDBC connection
    with patch.object(df.sql_ctx.sparkSession.sparkContext, "_jvm") as mock_jvm:
        mock_conn = MagicMock()
        mock_stmt = MagicMock()
        mock_jvm.java.sql.DriverManager.getConnection.return_value = mock_conn
        mock_conn.createStatement.return_value = mock_stmt

        transformer = FrameworkDDLExecutor(ddl=ddl, jdbc_url=jdbc_url)
        result_df = transformer.transform(df)

        mock_jvm.java.sql.DriverManager.getConnection.assert_called_once_with(jdbc_url)
        mock_conn.createStatement.assert_called_once()
        mock_stmt.execute.assert_called_once()
        mock_stmt.close.assert_called_once()
        mock_conn.close.assert_called_once()
        assert result_df is df



import json
from typing import Any, Dict, Iterable, List, Optional, Union, Generator

from helix_fhir_client_sdk.responses.fhir_merge_response import FhirMergeResponse
from pyspark.sql.types import Row

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.register import register
from spark_pipeline_framework.utilities.fhir_helpers.fhir_sender_operation import (
    FhirSenderOperation,
)
from spark_pipeline_framework.utilities.fhir_helpers.fhir_merge_response_item_schema import (
    FhirMergeResponseItemSchema,
)
from spark_pipeline_framework.transformers.fhir_sender.v1.fhir_sender_helpers import (
    send_fhir_delete,
    send_json_bundle_to_fhir,
    update_json_bundle_to_fhir,
)
from spark_pipeline_framework.utilities.json_helpers import convert_dict_to_fhir_json


class FhirSenderProcessor:
    @staticmethod
    # function that is called for each partition
    def send_partition_to_server(
        *,
        partition_index: int,
        rows: Iterable[Row],
        desired_partitions: int,
        operation: Union[FhirSenderOperation, str],
        server_url: str,
        resource_name: str,
        name: Optional[str],
        auth_server_url: Optional[str],
        auth_client_id: Optional[str],
        auth_client_secret: Optional[str],
        auth_login_token: Optional[str],
        auth_scopes: Optional[List[str]],
        auth_access_token: Optional[str],
        additional_request_headers: Optional[Dict[str, str]],
        log_level: Optional[str],
        batch_size: Optional[int],
        validation_server_url: Optional[str],
        retry_count: Optional[int],
        exclude_status_codes_from_retry: Optional[List[int]],
        smart_merge: Optional[bool] = None,
    ) -> Generator[List[Dict[str, Any]], None, None]:
        """
        This function processes a partition

        This has to be a static function to avoid creating a closure around a class
        https://spark.apache.org/docs/latest/rdd-programming-guide.html#passing-functions-to-spark


        """
        register()

        json_data_list: List[Row] = list(rows)
        logger = get_logger(__name__)
        assert server_url

        if len(json_data_list) == 0:
            yield []
        print(
            f"Sending batch {partition_index}/{desired_partitions} "
            f"containing {len(json_data_list)} rows "
            f"for operation {operation} "
            f"to {server_url}/{resource_name}. [{name}].."
        )
        request_id_list: List[str] = []
        responses: List[Dict[str, Any]] = []
        if FhirSenderOperation.operation_equals(
            operation, FhirSenderOperation.FHIR_OPERATION_PATCH
        ):
            for item in json_data_list:
                item_value = json.loads(item["value"])
                payload = [
                    json.loads(convert_dict_to_fhir_json(payload_item))
                    for payload_item in item_value["payload"]
                ]
                patch_result: Optional[Dict[str, Any]] = update_json_bundle_to_fhir(
                    obj_id=item_value["id"],
                    json_data=json.dumps(payload),
                    server_url=server_url,
                    operation=operation,
                    validation_server_url=validation_server_url,
                    resource=resource_name,
                    logger=logger,
                    auth_server_url=auth_server_url,
                    auth_client_id=auth_client_id,
                    auth_client_secret=auth_client_secret,
                    auth_login_token=auth_login_token,
                    auth_scopes=auth_scopes,
                    auth_access_token=auth_access_token,
                    additional_request_headers=additional_request_headers,
                    log_level=log_level,
                )
                if patch_result:
                    responses.append(patch_result)
        elif FhirSenderOperation.operation_equals(
            operation, FhirSenderOperation.FHIR_OPERATION_PUT
        ):
            for item in json_data_list:
                item_value = json.loads(item["value"])
                put_result: Optional[Dict[str, Any]] = update_json_bundle_to_fhir(
                    obj_id=item_value["id"],
                    json_data=convert_dict_to_fhir_json(item_value),
                    server_url=server_url,
                    operation=operation,
                    validation_server_url=validation_server_url,
                    resource=resource_name,
                    logger=logger,
                    auth_server_url=auth_server_url,
                    auth_client_id=auth_client_id,
                    auth_client_secret=auth_client_secret,
                    auth_login_token=auth_login_token,
                    auth_scopes=auth_scopes,
                    auth_access_token=auth_access_token,
                    additional_request_headers=additional_request_headers,
                    log_level=log_level,
                )
                if put_result:
                    responses.append(put_result)
        elif FhirSenderOperation.operation_equals(
            operation, FhirSenderOperation.FHIR_OPERATION_DELETE
        ):
            # FHIR doesn't support bulk deletes, so we have to send one at a time
            responses = [
                send_fhir_delete(
                    obj_id=(item if "id" in item else json.loads(item["value"]))[
                        "id"
                    ],  # parse the JSON and extract the id
                    server_url=server_url,
                    resource=resource_name,
                    logger=logger,
                    auth_server_url=auth_server_url,
                    auth_client_id=auth_client_id,
                    auth_client_secret=auth_client_secret,
                    auth_login_token=auth_login_token,
                    auth_scopes=auth_scopes,
                    auth_access_token=auth_access_token,
                    additional_request_headers=additional_request_headers,
                    log_level=log_level,
                )
                for item in json_data_list
            ]
        elif FhirSenderOperation.operation_equals(
            operation, FhirSenderOperation.FHIR_OPERATION_MERGE
        ):
            if batch_size == 1:
                # ensure we call one at a time. Partitioning does not guarantee that each
                #   partition will have exactly the batch size
                auth_access_token1: Optional[str] = auth_access_token
                i: int = 0
                count: int = len(json_data_list)
                for item in json_data_list:
                    i += 1
                    current_bundle: List[Dict[str, Any]]
                    json_data_list_1: List[str] = [
                        (
                            convert_dict_to_fhir_json(item.asDict(recursive=True))
                            if "id" in item
                            else item["value"]
                        )
                    ]
                    first_item: Optional[Dict[str, Any]] = (
                        json.loads(json_data_list_1[0])
                        if len(json_data_list_1) > 0
                        else None
                    )
                    first_id = first_item.get("id") if first_item is not None else None
                    print(
                        f"Sending {i} of {count} from partition {partition_index} "
                        f"for {resource_name}: {first_id}"
                    )
                    if first_id:
                        result: Optional[FhirMergeResponse] = send_json_bundle_to_fhir(
                            id_=first_id,
                            json_data_list=json_data_list_1,
                            server_url=server_url,
                            validation_server_url=validation_server_url,
                            resource=resource_name,
                            logger=logger,
                            auth_server_url=auth_server_url,
                            auth_client_id=auth_client_id,
                            auth_client_secret=auth_client_secret,
                            auth_login_token=auth_login_token,
                            auth_scopes=auth_scopes,
                            auth_access_token=auth_access_token1,
                            additional_request_headers=additional_request_headers,
                            log_level=log_level,
                            retry_count=retry_count,
                            exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                            smart_merge=smart_merge,
                        )
                        if result:
                            auth_access_token1 = result.access_token
                            if result.request_id:
                                request_id_list.append(result.request_id)
                            responses.extend(result.responses)
            else:
                # send a whole batch to the server at once
                json_data_list_1 = [
                    (
                        convert_dict_to_fhir_json(item.asDict(recursive=True))
                        if "id" in item
                        else item["value"]
                    )
                    for item in json_data_list
                ]
                first_item = (
                    json.loads(json_data_list_1[0])
                    if len(json_data_list_1) > 0
                    else None
                )
                first_id = first_item.get("id") if first_item is not None else None
                if first_id:
                    result = send_json_bundle_to_fhir(
                        id_=first_id,
                        json_data_list=json_data_list_1,
                        server_url=server_url,
                        validation_server_url=validation_server_url,
                        resource=resource_name,
                        auth_server_url=auth_server_url,
                        auth_client_id=auth_client_id,
                        auth_client_secret=auth_client_secret,
                        auth_login_token=auth_login_token,
                        auth_scopes=auth_scopes,
                        auth_access_token=auth_access_token,
                        additional_request_headers=additional_request_headers,
                        logger=logger,
                        log_level=log_level,
                        retry_count=retry_count,
                        exclude_status_codes_from_retry=exclude_status_codes_from_retry,
                    )
                    if result and result.request_id:
                        request_id_list.append(result.request_id)
                    if result:
                        responses = result.responses
        # each item in responses is either a json object
        #   or a list of json objects
        error_count: int = 0
        updated_count: int = 0
        created_count: int = 0
        deleted_count: int = 0
        errors: List[str] = []
        response: Dict[str, Any]
        for response in responses:
            if response.get(FhirMergeResponseItemSchema.updated) is True:
                updated_count += 1
            if response.get(FhirMergeResponseItemSchema.created) is True:
                created_count += 1
            if response.get(FhirMergeResponseItemSchema.deleted) is True:
                deleted_count += 1
            if response.get(FhirMergeResponseItemSchema.issue) is not None:
                error_count += 1
                errors.append(json.dumps(response[FhirMergeResponseItemSchema.issue]))
        print(
            f"Received response for batch {partition_index}/{desired_partitions} "
            f"request_ids:[{', '.join(request_id_list)}] "
            f"total={len(json_data_list)}, error={error_count}, "
            f"created={created_count}, updated={updated_count}, deleted={deleted_count} "
            f"to {server_url}/{resource_name}. "
            f"[{name}] "
            f"Response={json.dumps(responses, default=str)}"
        )
        if len(errors) > 0:
            logger.error(
                f"---- errors for {partition_index}/{desired_partitions} "
                f"to {server_url}/{resource_name} -----"
            )
            for error in errors:
                logger.error(error)
            logger.error("---- end errors -----")

        yield responses



import dataclasses
import json
import os
import socket
import threading
from typing import Optional, Dict, Any

from pyspark import TaskContext
from pyspark.sql.types import (
    StructType,
    StructField,
    IntegerType,
    StringType,
    BooleanType,
)


@dataclasses.dataclass
class SparkPartitionInformation:
    partition_index: Optional[int]
    chunk_index: Optional[int]
    process_id: int
    thread_name: str
    stage_id: Optional[int]
    task_attempt_id: Optional[int]
    cpus: Optional[int]
    is_driver: bool
    host_name: Optional[str]

    @classmethod
    def from_current_task_context(
        cls, *, chunk_index: Optional[int] = None
    ) -> "SparkPartitionInformation":
        """
        Create a SparkPartitionInformation object from the current TaskContext.

        :param chunk_index: The index of the chunk within the partition.
        :return: A SparkPartitionInformation
        """
        # https://spark.apache.org/docs/3.5.1/api/python/reference/api/pyspark.TaskContext.html
        return cls.from_task_context(
            task_context=TaskContext.get(), chunk_index=chunk_index
        )

    @classmethod
    def from_task_context(
        cls, *, task_context: TaskContext | None, chunk_index: Optional[int]
    ) -> "SparkPartitionInformation":
        """
        Create a SparkPartitionInformation object from a TaskContext.

        :param task_context: The TaskContext.
        :param chunk_index: The index of the chunk within the partition.
        :return: A SparkPartitionInformation
        """
        # https://spark.apache.org/docs/3.5.1/api/python/reference/api/pyspark.TaskContext.html
        return cls(
            partition_index=(
                task_context.partitionId() if task_context is not None else None
            ),
            chunk_index=chunk_index,
            process_id=os.getpid(),
            thread_name=threading.current_thread().name,
            stage_id=task_context.stageId() if task_context is not None else None,
            task_attempt_id=(
                task_context.taskAttemptId() if task_context is not None else None
            ),
            cpus=task_context.cpus() if task_context is not None else None,
            is_driver=task_context is None,
            host_name=os.getenv("HOSTNAME") or socket.gethostname(),
        )

    def __str__(self) -> str:
        """
        Provide a string representation of the object.

        :return: A string representation of the object.
        """
        return (
            f" | Partition: {self.partition_index}"
            + (f" | Chunk: {self.chunk_index}" if self.chunk_index is not None else "")
            + (f" | Host: {self.host_name}" if self.host_name is not None else "")
            + f" | Process: {self.process_id}"
            + f" | Thread: {self.thread_name} ({threading.get_ident()})"
            + (f" | Spark Driver: {self.is_driver}" if self.is_driver else "")
            + f" | Spark Stage Id: {self.stage_id}"
            + f" | Spark Task Attempt ID: {self.task_attempt_id}"
            + f" | Spark CPUs: {self.cpus}"
        )

    def to_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    @staticmethod
    def get_schema() -> StructType:
        return StructType(
            [
                StructField("partition_index", IntegerType(), True),
                StructField("chunk_index", IntegerType(), True),
                StructField("process_id", IntegerType(), False),
                StructField("thread_name", StringType(), False),
                StructField("stage_id", IntegerType(), True),
                StructField("task_attempt_id", IntegerType(), True),
                StructField("cpus", IntegerType(), True),
                StructField("is_driver", BooleanType(), False),
                StructField("host_name", StringType(), True),
            ]
        )

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_string: str) -> "SparkPartitionInformation":
        return cls(**json.loads(json_string))

    @classmethod
    def from_dict(cls, dictionary: Dict[str, Any]) -> "SparkPartitionInformation":
        return cls(**dictionary)



from typing import Any
from typing import Optional


class BaseSlackClient:
    def post_message_to_slack(
        self,
        text: str,
        blocks: Any = None,
        use_conversation_threads: bool = True,
        slack_thread: Optional[str] = None,
    ) -> Any:
        raise NotImplementedError



import json
from pathlib import Path
from typing import Any, Dict, List, Optional, cast

from mockserver_client.mock_requests_loader import (
    load_mock_source_api_responses_from_folder,
    load_mock_source_api_json_responses,
)
from mockserver_client.mockserver_client import MockServerFriendlyClient
from pyspark.sql import SparkSession, DataFrame
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.http_data_receiver.v2.http_data_receiver import (
    HttpDataReceiver,
)
from spark_pipeline_framework.utilities.api_helper.http_request import HelixHttpRequest
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


def test_http_data_receiver(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")

    test_name = "http_data_receiver_test"
    mock_server_url = "http://mock-server:1080"
    mock_client = MockServerFriendlyClient(mock_server_url)
    mock_client.clear(f"/{test_name}/*.*")
    load_mock_source_api_responses_from_folder(
        folder=data_dir.joinpath("response"),
        mock_client=mock_client,
        url_prefix=test_name,
    )

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    # Act
    request = HelixHttpRequest(url=f"{mock_server_url}/{test_name}{'/res1.json'}")
    with ProgressLogger() as progress_logger:
        HttpDataReceiver(
            http_request=request, view_name=test_name, progress_logger=progress_logger
        ).transform(df)

    # Assert
    result_df: DataFrame = spark_session.table(test_name)
    result_df.printSchema()
    result_df.show(truncate=False)
    result: List[Dict[str, Dict[str, List[int]]]] = (
        result_df.toJSON()
        .map(lambda j: cast(Dict[str, Dict[str, List[int]]], json.loads(j)))
        .collect()
    )
    result_: Dict[str, Dict[str, List[int]]] = result[0]
    assert result_.get("result") is not None
    assert result_.get("result").get("sub") == [1, 2, 3]  # type: ignore


def test_http_paging(spark_session: SparkSession) -> None:
    # Arrange
    def next_request(
        http_request: HelixHttpRequest,
        response: Dict[str, Any],
        progress_logger: Optional[ProgressLogger],
    ) -> Optional[HelixHttpRequest]:
        query_string: Dict[str, List[str]] = http_request.get_querystring()
        page = int(query_string["page"][0])
        if page >= response["result"]["total"]:
            return None
        else:
            return HelixHttpRequest(
                url=f"{mock_server_url}/{test_name}?page={page + 1}"
            )

    data_dir: Path = Path(__file__).parent.joinpath("./")

    test_name = "test_http_paging"
    mock_server_url = "http://mock-server:1080"
    mock_client = MockServerFriendlyClient(mock_server_url)
    mock_client.clear(f"/{test_name}/*.*")
    load_mock_source_api_json_responses(
        folder=data_dir.joinpath("pages"),
        mock_client=mock_client,
        url_prefix=test_name,
    )

    df: DataFrame = create_empty_dataframe(spark_session=spark_session)

    # Act
    request = HelixHttpRequest(url=f"{mock_server_url}/{test_name}{'?page=1'}")
    with ProgressLogger() as progress_logger:
        HttpDataReceiver(
            http_request=request,
            view_name=test_name,
            progress_logger=progress_logger,
            next_request_generator=next_request,
        ).transform(df)

    # Assert
    result_df: DataFrame = spark_session.table(test_name)
    result_df.printSchema()
    result_df.show(truncate=False)
    result = (
        result_df.toJSON().map(lambda j: cast(Dict[str, Any], json.loads(j))).collect()
    )
    assert result[0].get("result") == {"page": 1, "total": 2}
    assert result[1].get("result") == {"page": 2, "total": 2}
    assert len(result) == 2



from typing import Dict, Any, Optional, Callable

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class FrameworkRunFunctionTransformer(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        fn: Callable[[DataFrame], DataFrame],
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ) -> None:
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        # add a param
        self.fn: Param[Callable[[DataFrame], DataFrame]] = Param(self, "fn", "")
        self._setDefault(fn=fn)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        fn: Callable[[DataFrame], DataFrame] = self.getFunction()

        df = fn(df)

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFunction(self) -> Callable[[DataFrame], DataFrame]:
        return self.getOrDefault(self.fn)



import re
from typing import Any
from pyspark.sql.dataframe import DataFrame
from pyathena import connect as athena_connect
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.utilities.athena.athena_source_file_type import (
    AthenaSourceFileType,
)


class Athena:
    # S3_STAGING_PATH = "s3://bwell-helix-log-scratchpad-ue1/athena_query_results/"

    @staticmethod
    def drop_create_table(
        table_name: str,
        schema_name: str,
        df: DataFrame,
        s3_source_path: str,
        source_file_type: AthenaSourceFileType,
        s3_temp_folder: str,
    ) -> None:
        drop_ddl: str = Athena._generate_athena_drop_table_ddl(table_name)
        Athena._execute_ddl(schema_name, drop_ddl, s3_temp_folder=s3_temp_folder)
        create_ddl: str = Athena._generate_athena_create_table_ddl(
            df, table_name, s3_source_path, source_file_type
        )
        Athena._execute_ddl(schema_name, create_ddl, s3_temp_folder=s3_temp_folder)

    @staticmethod
    def _generate_athena_drop_table_ddl(athena_table_name: str) -> str:
        return f"DROP TABLE IF EXISTS `{athena_table_name}`"

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    @staticmethod
    def _generate_athena_create_table_ddl(
        df: DataFrame,
        athena_table_name: str,
        s3_location: str,
        source_file_type: AthenaSourceFileType,
    ) -> str:
        s3_location = s3_location.replace("s3a://", "s3://")
        ddl = []
        fields = df.schema

        ddl.append(f"CREATE EXTERNAL TABLE IF NOT EXISTS `{athena_table_name}` (")

        for field in fields.fieldNames()[:-1]:
            ddl.append(
                f"\t`{fields[field].name}` {Athena._get_ddl_data_type(fields[field])},"
            )
        last = fields[fields.fieldNames()[-1]]
        ddl.append(f"\t`{last.name}` {Athena._get_ddl_data_type(last)}")
        ddl.append(")")

        if source_file_type == AthenaSourceFileType.CSV_QUOTED:
            ddl.append("ROW FORMAT SERDE 'org.apache.hadoop.hive.serde2.OpenCSVSerde'")
            ddl.append("WITH SERDEPROPERTIES (")
            ddl.append("  'separatorChar' = ',',")
            ddl.append("  'quoteChar' = '\"',")
            ddl.append("  'escapeChar' = '\\\\'")
            ddl.append("  )")
        elif source_file_type == AthenaSourceFileType.CSV_UNQUOTED:
            ddl.append("ROW FORMAT DELIMITED")
            ddl.append("  FIELDS TERMINATED BY ','")
            ddl.append("  ESCAPED BY '\\\\'")
            ddl.append("  LINES TERMINATED BY '\\n'")
        elif source_file_type == AthenaSourceFileType.JSONL:
            ddl.append("ROW FORMAT SERDE 'org.openx.data.jsonserde.JsonSerDe'")
        elif source_file_type == AthenaSourceFileType.PARQUET:
            ddl.append("STORED AS PARQUET")

        ddl.append(f"LOCATION '{s3_location}'")

        if source_file_type == AthenaSourceFileType.PARQUET:
            ddl.append('tblproperties ("parquet.compression"="SNAPPY")')

        return "\n".join(ddl)

    @staticmethod
    def _get_ddl_data_type(field: Any) -> str:
        # wrap all field names in backticks to avoid reserved word collisions
        return str(
            re.sub("(\\w*?)?:(\\w*?)", "`\\1`:\\2", field.dataType.simpleString())
        )

    @staticmethod
    def _execute_ddl(schema_name: str, ddl: str, s3_temp_folder: str) -> None:
        try:
            cursor = athena_connect(
                schema_name=schema_name, s3_staging_dir=s3_temp_folder
            ).cursor()
            cursor.execute(ddl)
        except Exception as e:
            logger = get_logger(__name__)
            logger.exception(
                f"ERROR executing Athena query on schema: {schema_name} using temp_folder:{s3_temp_folder}"
            )
            logger.error("---------- Query ------------")
            logger.error(ddl)
            logger.error("------------------------------")
            logger.error(e)



from dataclasses import dataclass
from datetime import datetime


@dataclass
class AsyncPandasBatchFunctionRunContext:
    """This class is used to pass the context"""

    """ The index of the partition"""
    partition_index: int
    """ The index of the chunk"""
    chunk_index: int
    """ The range of indices of the input values within the chunk."""
    chunk_input_range: range
    """ The start time of the partition processing """
    partition_start_time: datetime



from typing import List, Set, Tuple, Type, Optional, Any

import structlog
from motor.motor_asyncio import (
    AsyncIOMotorClient,
    AsyncIOMotorCollection,
    AsyncIOMotorCursor,
)
from pymongo import UpdateOne

from spark_pipeline_framework.utilities.document_db_connection.v1.document_db_connection import (
    DocumentDbServerUrl,
)

from spark_pipeline_framework.utilities.helix_geolocation.v2.cache.cache_handler import (
    CacheHandler,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.cache.cache_result import (
    CacheResult,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.vendor_response import (
    VendorResponse,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.base_vendor_api_response import (
    BaseVendorApiResponse,
)

logger = structlog.get_logger(__file__)


class DocumentDBCacheHandler(CacheHandler):
    def __init__(
        self,
        database_name: str = "helix_address_cache",
        collection_name: str = "helix_address_cache",
        server_url: Optional[str] = None,
        collection: Optional[AsyncIOMotorCollection[Any]] = None,
    ):
        self.__server_url: Optional[str] = server_url
        self.database_name: str = database_name
        self.collection_name: str = collection_name
        self.__collection: Optional[AsyncIOMotorCollection[Any]] = collection

    @property
    async def _server_url(self) -> str:
        if self.__server_url is not None:
            return self.__server_url
        else:
            self.__server_url = DocumentDbServerUrl().get_server_url()
            return self.__server_url

    @property
    async def _collection(self) -> AsyncIOMotorCollection[Any]:
        if self.__collection is not None:
            return self.__collection
        else:
            mongo_client: AsyncIOMotorClient[Any] = AsyncIOMotorClient(
                await self._server_url
            )
            self._db = mongo_client.get_database(self.database_name)
            self.__collection = self._db[self.collection_name]
            await self.__collection.create_index("address_hash", unique=True)
            return self.__collection

    async def check_cache(self, raw_addresses: List[RawAddress]) -> CacheResult:
        unique_address_hashes: Set[str] = set([a.to_hash() for a in raw_addresses])
        query = {"address_hash": {"$in": list(unique_address_hashes)}}
        collection: AsyncIOMotorCollection[Any] = await self._collection
        cursor: AsyncIOMotorCursor[Any] = collection.find(query)

        found_vendor_response: List[VendorResponse[BaseVendorApiResponse]] = []
        async for r in cursor:
            matching_raw: List[RawAddress] = [
                raw for raw in raw_addresses if r["address_hash"] == raw.to_hash()
            ]
            vendor_name = r["vendor_name"]
            vendor_class: Type[StandardizingVendor[BaseVendorApiResponse]] = (
                self._get_vendor_class(vendor_name)
            )
            api_response_class: Type[BaseVendorApiResponse] = (
                vendor_class.get_api_response_class()
            )
            assert (
                api_response_class is not None
            ), f"api_response_class is None for {vendor_name}"
            found_vendor_response.extend(
                [
                    VendorResponse(
                        api_call_response=api_response_class.from_dict(
                            r["vendor_std_address"]
                        ),
                        related_raw_address=raw,
                        vendor_name=vendor_name,
                        response_version=r["response_version"],
                    )
                    for raw in matching_raw
                ]
            )

        found_std_addr: List[StandardizedAddress]
        found_ids: List[str]
        found_std_addr, found_ids = self._convert_to_std_address(found_vendor_response)
        not_found: List[RawAddress] = [
            a for a in raw_addresses if a.get_id() not in found_ids
        ]
        return CacheResult(found=found_std_addr, not_found=not_found)

    async def save_to_cache(self, vendor_responses: List[VendorResponse[Any]]) -> None:
        requests = [
            UpdateOne(
                {"address_hash": vr.related_raw_address.to_hash()},
                {
                    "$set": {
                        "vendor_std_address": vr.api_call_response.to_dict(),
                        "vendor_name": vr.vendor_name,
                        "response_version": vr.response_version,
                    }
                },
                upsert=True,
            )
            for vr in vendor_responses
            if vr.related_raw_address
        ]
        if requests:
            collection = await self._collection
            write_result = await collection.bulk_write(requests=requests)
            logger.info("Cache write result:", result=write_result.bulk_api_result)
        else:
            logger.warning(
                "No responses to write to cache!",
                number_of_vendor_responses=len(vendor_responses),
                requests=len(requests),
            )

    def _get_vendor_class(
        self, vendor_name: str
    ) -> Type[StandardizingVendor[BaseVendorApiResponse]]:
        return super()._get_vendor_class(vendor_name)

    @staticmethod
    def _convert_to_std_address(
        vendor_responses: List[VendorResponse[BaseVendorApiResponse]],
    ) -> Tuple[List[StandardizedAddress], List[str]]:
        std_addresses: List[StandardizedAddress] = []
        found_ids: List[str] = []
        vr: VendorResponse[BaseVendorApiResponse]
        for vr in vendor_responses:
            if vr.related_raw_address and vr.related_raw_address.get_id():
                standardized_address = vr.api_call_response.to_standardized_address(
                    address_id=vr.related_raw_address.get_id()
                )
                if standardized_address:
                    std_addresses.append(standardized_address)
                if vr.related_raw_address:
                    address_id: Optional[str] = vr.related_raw_address.get_id()
                    assert address_id is not None
                    found_ids.append(address_id)
        return std_addresses, found_ids



import pytest
from aioresponses import aioresponses
from typing import List, Dict, Any

from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendor_response_key_error import (
    VendorResponseKeyError,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.melissa_standardizing_vendor import (
    MelissaStandardizingVendor,
)


@pytest.mark.asyncio
async def test_standardize_async_with_valid_response() -> None:
    """
    Test standardize_async method with a valid response from Melissa API.
    """
    raw_addresses = [
        RawAddress(
            address_id="1",
            line1="123 Test St",
            city="Test City",
            state="CA",
            zipcode="12345",
            country="US",
        )
    ]
    expected_response = {
        "Records": [
            {
                "RecordID": "1",
                "FormattedAddress": "123 Test St, USA",
                "Locality": "Test City",
                "AdministrativeArea": "CA",
                "SubAdministrativeArea": "",
                "PostalCode": "12345",
                "CountryISO3166_1_Alpha2": "US",
                "Latitude": "37.774929",
                "Longitude": "-122.419416",
            }
        ]
    }

    vendor = MelissaStandardizingVendor(license_key="test_license_key")

    with aioresponses() as m:
        m.post(
            "https://address.melissadata.net/v3/WEB/GlobalAddress/doglobaladdress",
            payload=expected_response,
        )

        result = await vendor.standardize_async(raw_addresses)

    assert len(result) == 1, f"Result: {result}"
    assert result[0].api_call_response.RecordID == "1", f"Result: {result}"
    assert result[0].related_raw_address.get_id() == "1", f"Result: {result}"


@pytest.mark.asyncio
async def test_standardize_async_with_empty_response() -> None:
    """
    Test standardize_async method when Melissa API returns an empty response.
    """
    raw_addresses = [
        RawAddress(
            address_id="1",
            line1="123 Test St",
            city="Test City",
            state="CA",
            zipcode="12345",
            country="US",
        )
    ]
    expected_response: Dict[str, List[Any]] = {"Records": []}

    vendor = MelissaStandardizingVendor(license_key="test_license_key")

    with aioresponses() as m:
        m.post(
            "https://address.melissadata.net/v3/WEB/GlobalAddress/doglobaladdress",
            payload=expected_response,
        )

        result = await vendor.standardize_async(raw_addresses)

    assert len(result) == 0, f"Result: {result}"


@pytest.mark.asyncio
async def test_standardize_async_with_key_error() -> None:
    """
    Test standardize_async method when Melissa API returns an invalid response,
    causing a VendorResponseKeyError to be raised.
    """
    raw_addresses = [
        RawAddress(
            address_id="1",
            line1="123 Test St",
            city="Test City",
            state="CA",
            zipcode="12345",
            country="US",
        )
    ]
    expected_response = {"InvalidKey": "InvalidValue"}

    vendor = MelissaStandardizingVendor(license_key="test_license_key")

    with aioresponses() as m:
        m.post(
            "https://address.melissadata.net/v3/WEB/GlobalAddress/doglobaladdress",
            payload=expected_response,
        )

        with pytest.raises(VendorResponseKeyError):
            await vendor.standardize_async(raw_addresses)


@pytest.mark.asyncio
async def test_api_calls_limit() -> None:
    """
    Test standardize_async method to ensure API call limit is respected.
    """
    raw_addresses = [
        RawAddress(
            address_id="1",
            line1="123 Test St",
            city="Test City",
            state="CA",
            zipcode="12345",
            country="US",
        )
    ]
    expected_response = {
        "Records": [
            {
                "RecordID": "1",
                "FormattedAddress": "123 Test St, USA",
                "Locality": "Test City",
                "AdministrativeArea": "CA",
                "SubAdministrativeArea": "",
                "PostalCode": "12345",
                "CountryISO3166_1_Alpha2": "US",
                "Latitude": "37.774929",
                "Longitude": "-122.419416",
            }
        ]
    }

    vendor = MelissaStandardizingVendor(
        license_key="test_license_key", api_calls_limit=1
    )

    with aioresponses() as m:
        m.post(
            "https://address.melissadata.net/v3/WEB/GlobalAddress/doglobaladdress",
            payload=expected_response,
        )

        # First call should succeed
        result1 = await vendor.standardize_async(raw_addresses)
        assert len(result1) == 1, f"Result: {result1}"

        # Second call should respect the limit and return a mock response
        result2 = await vendor.standardize_async(raw_addresses)
        assert len(result2) == 1, f"Result: {result2}"
        assert result2[0].api_call_response.RecordID == "1", f"Result: {result2}"


@pytest.mark.asyncio
async def test_standardize_async_with_custom_api_call() -> None:
    """
    Test standardize_async method when using a custom API call function.
    """
    raw_addresses = [
        RawAddress(
            address_id="1",
            line1="123 Test St",
            city="Test City",
            state="CA",
            zipcode="12345",
            country="US",
        )
    ]

    async def custom_api_call(raw_addresses: List[RawAddress]) -> Dict[str, Any]:
        return {
            "Records": [
                {
                    "RecordID": "1",
                    "FormattedAddress": "123 Test St, USA",
                    "Locality": "Test City",
                    "AdministrativeArea": "CA",
                    "SubAdministrativeArea": "",
                    "PostalCode": "12345",
                    "CountryISO3166_1_Alpha2": "US",
                    "Latitude": "37.774929",
                    "Longitude": "-122.419416",
                }
            ]
        }

    vendor = MelissaStandardizingVendor(license_key="", custom_api_call=custom_api_call)

    result = await vendor.standardize_async(raw_addresses)

    assert len(result) == 1, f"Result: {result}"
    assert result[0].api_call_response.RecordID == "1", f"Result: {result}"
    assert result[0].related_raw_address.get_id() == "1", f"Result: {result}"
    assert (
        result[0].api_call_response.FormattedAddress == "123 Test St, USA"
    ), f"Result: {result}"



import asyncio
import dataclasses
from datetime import datetime
from logging import Logger
from typing import (
    List,
    Dict,
    Any,
    Optional,
    AsyncGenerator,
    Iterable,
    Tuple,
    Generator,
)

from pyspark.sql import SparkSession, DataFrame
from pyspark.sql.types import StructType, StructField, StringType

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_batch_function_run_context import (
    AsyncPandasBatchFunctionRunContext,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_dataframe_udf import (
    AsyncPandasDataFrameUDF,
)
from spark_pipeline_framework.utilities.spark_partition_information.v1.spark_partition_information import (
    SparkPartitionInformation,
)


def test_async_pandas_dataframe_udf_large(spark_session: SparkSession) -> None:
    print()
    df: DataFrame = spark_session.createDataFrame(
        [
            ("00100000000", "Qureshi"),
            ("00200000000", "Vidal"),
            ("00300000000", "Smith"),
            ("00400000000", "Johnson"),
            ("00500000000", "Williams"),
            ("00600000000", "Brown"),
            ("00700000000", "Jones"),
            ("00800000000", "Garcia"),
            ("00900000000", "Miller"),
            ("01000000000", "Davis"),
            ("01100000000", "Rodriguez"),
            ("01200000000", "Martinez"),
            ("01300000000", "Hernandez"),
            ("01400000000", "Lopez"),
            ("01500000000", "Gonzalez"),
            ("01600000000", "Wilson"),
            ("01700000000", "Anderson"),
            ("01800000000", "Thomas"),
            ("01900000000", "Taylor"),
            ("02000000000", "Moore"),
        ],
        ["id", "name"],
    )

    print(f"Partition Count: {df.rdd.getNumPartitions()}")

    # Function to list partition IDs and column values
    def list_partition_ids_and_values(
        partition_index: int, iterator: Iterable[Any]
    ) -> Generator[Tuple[int, List[str]], None, None]:
        for row in iterator:
            yield partition_index, row["id"]

    partition_id_and_values: List[Tuple[int, List[str]]] = (
        df.rdd.mapPartitionsWithIndex(list_partition_ids_and_values).collect()
    )
    print(f"Partition IDs: {partition_id_and_values}")

    @dataclasses.dataclass
    class MyParameters:
        pandas_udf_parameters: AsyncPandasUdfParameters
        log_level: str = "INFO"

    # noinspection PyUnusedLocal
    async def test_async(
        run_context: AsyncPandasBatchFunctionRunContext,
        input_values: List[Dict[str, Any]],
        parameters: Optional[MyParameters],
        additional_parameters: Optional[Dict[str, Any]],
    ) -> AsyncGenerator[Dict[str, Any], None]:

        logger: Logger = get_logger(
            __name__,
            level=(
                parameters.log_level if parameters and parameters.log_level else "INFO"
            ),
        )

        spark_partition_information: SparkPartitionInformation = (
            SparkPartitionInformation.from_current_task_context(
                chunk_index=run_context.chunk_index,
            )
        )
        if parameters is not None and parameters.log_level == "DEBUG":
            ids = [input_value["id"] for input_value in input_values]
            message: str = f"In test_async"
            # Get the current time
            current_time = datetime.now()

            # Format the time to include hours, minutes, seconds, and milliseconds
            formatted_time = current_time.strftime("%H:%M:%S.%f")[:-3]
            formatted_message: str = (
                f"{formatted_time}: "
                f"{message}"
                f" | Partition: {run_context.partition_index}"
                f" | Chunk: {run_context.chunk_index}"
                f" | range: {run_context.chunk_input_range.start}-{run_context.chunk_input_range.stop}"
                f" | Ids ({len(ids)}): {ids}"
                f" | {spark_partition_information}"
            )
            # print(f"print: {formatted_message}")
            logger.debug(f"logger: {formatted_message}")

        input_value: Dict[str, Any]
        for input_value in input_values:
            # simulate a delay
            await asyncio.sleep(0.1)
            yield {
                "id": input_value["id"],
                "name": input_value["name"] + "_processed",
                "partition_info": spark_partition_information.to_json(),
            }

    output_schema: StructType = StructType(
        [
            StructField("id", StringType(), False),
            StructField("name", StringType(), False),
            StructField("partition_info", StringType(), False),
        ]
    )

    result_df: DataFrame = df.mapInPandas(
        AsyncPandasDataFrameUDF(
            parameters=MyParameters(
                log_level="DEBUG", pandas_udf_parameters=AsyncPandasUdfParameters()
            ),
            async_func=test_async,  # type: ignore[arg-type]
            pandas_udf_parameters=AsyncPandasUdfParameters(max_chunk_size=2),
        ).get_pandas_udf(),
        schema=output_schema,
    )

    print("result_df")
    result_df.show(truncate=False)

    assert result_df.count() == 20
    assert result_df.collect()[0]["name"] == "Qureshi_processed"
    assert result_df.collect()[1]["name"] == "Vidal_processed"
    assert result_df.collect()[0]["partition_info"] != ""
